"""Datapipe to check variable and dimension names exist"""
from typing import Iterable, Optional, Union

import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("check_vars_and_dims")
class CheckVarsAndDimsIterDataPipe(IterDataPipe):
    """Check that the variables and dims exist"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        expected_dimensions: Optional[Iterable[str]] = None,
        expected_data_vars: Optional[Iterable[str]] = None,
        dataset_name: Optional[str] = None,
    ):
        """
        Checks data vars and dimensions for validation

        Args:
            source_datapipe: Source datapipe emitting a xarray dataset/dataarray
            expected_dimensions: The expected dimensions in the data
            expected_data_vars: The expected data variable names
            dataset_name: Name of subdataset to check, None for checking all of them
        """
        self.source_datapipe = source_datapipe
        self.expected_dimensions = expected_dimensions
        self.expected_data_vars = expected_data_vars
        self.dataset_name = dataset_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """
        Validate the data and return the validated xarray dataset

        Returns:
            Xarray dataset

        """
        for xr_data in self.source_datapipe:
            if self.dataset_name is None:
                if self.expected_data_vars is not None:
                    xr_data = validate_data_vars(xr_data, self.expected_data_vars)
                if self.expected_dimensions is not None:
                    xr_data = validate_dims(xr_data, self.expected_dimensions)
                    xr_data = validate_coords(xr_data, self.expected_dimensions)
            else:
                if self.expected_data_vars is not None:
                    xr_data[self.dataset_name] = validate_data_vars(
                        xr_data[self.dataset_name], self.expected_data_vars
                    )
                if self.expected_dimensions is not None:
                    xr_data[self.dataset_name] = validate_dims(
                        xr_data[self.dataset_name], self.expected_dimensions
                    )
                    xr_data[self.dataset_name] = validate_coords(
                        xr_data[self.dataset_name], self.expected_dimensions
                    )
            yield xr_data


def validate_data_vars(
    xr_data: Union[xr.DataArray, xr.Dataset], expected_data_vars: Iterable[str]
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Validate the data variables in the Xarray dataset

    Args:
        xr_data: Xarray dataset
        expected_data_vars: The expected data variable names

    Returns:
        Original Xarray dataset
    """
    data_var_names = xr_data.data_vars
    for data_var in expected_data_vars:
        assert data_var in data_var_names, f"{data_var} is not in all data_vars ({data_var_names})!"
    return xr_data


def validate_coords(
    xr_data: Union[xr.DataArray, xr.Dataset], expected_dimensions: Iterable[str]
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Validates that the coordinates exist and are not 0-length in the dataset

    Args:
        xr_data: Xarray dataset
        expected_dimensions: The expected dimensions

    Returns:
        Original Xarray dataset
    """
    for dim in expected_dimensions:
        coord = xr_data.coords[f"{dim}"]
        assert len(coord) > 0, f"{dim} is empty!"
    return xr_data


def validate_dims(
    xr_data: Union[xr.DataArray, xr.Dataset], expected_dimensions: Iterable[str]
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Validate the dimensions exist

    Args:
        xr_data: Xarray dataset to check
        expected_dimensions: The expected dimension names

    Returns:
        The original Xarray Dataset
    """
    assert all(dim in expected_dimensions for dim in xr_data.dims if dim != "example"), (
        f"dims is wrong! "
        f"dims is {xr_data.dims}. "
        f"But we expected {expected_dimensions}."
        " Note that 'example' is ignored, and the order is ignored"
    )
    return xr_data



"""Convert point PV sites to image output"""
import logging
from collections import defaultdict
from typing import Union

import numpy as np
import pandas as pd
import pvlib
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils import Zipper
from ocf_datapipes.utils.consts import Location
from ocf_datapipes.utils.geospatial import load_geostationary_area_definition_and_transform_osgb

logger = logging.getLogger(__name__)


@functional_datapipe("create_pv_image")
class CreatePVImageIterDataPipe(IterDataPipe):
    """Create PV image from individual sites"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        image_datapipe: IterDataPipe,
        normalize: bool = False,
        image_dim: str = "geostationary",
        max_num_pv_systems: int = -1,
        always_return_first: bool = False,
        seed: int = None,
        take_n_pv_values_per_pixel: int = -1,
        normalize_by_pvlib: bool = False,
    ):
        """
        Creates a 3D data cube of PV output image x number of timesteps

        TODO Include PV System IDs or something, currently just takes the outputs

        Args:
            source_datapipe: Source datapipe of PV data
            image_datapipe: Datapipe emitting images to get the shape from, with coordinates
            normalize: Whether to normalize based off the image max, or leave raw data
            image_dim: Dimension name for the x and y dimensions
            max_num_pv_systems: Max number of PV systems to use
            always_return_first: Always return the first image data cube, to save computation
                Only use for if making the image at the beginning of the stack
            seed: Random seed to use if using max_num_pv_systems
            take_n_pv_values_per_pixel: Take N PV values as the value for the pixel
                If -1, averages the PV generation if there are multiple PVs per pixel.
            normalize_by_pvlib: Normalize by pvlib's poa_global based off
                tilt/orientation/capacity/lat/lon of the system

        """
        if normalize_by_pvlib is not False or normalize is not False:
            assert normalize != normalize_by_pvlib, ValueError(
                "Cannot normalize by both max, and pvlib"
            )
        self.source_datapipe = source_datapipe
        self.image_datapipe = image_datapipe
        self.normalize = normalize
        self.x_dim = "x_" + image_dim
        self.y_dim = "y_" + image_dim
        self.max_num_pv_systems = max_num_pv_systems
        self.rng = np.random.default_rng(seed=seed)
        self.always_return_first = always_return_first
        self.take_n_pv_values_per_pixel = take_n_pv_values_per_pixel
        self.normalize_by_pvlib = normalize_by_pvlib

    def __iter__(self) -> xr.DataArray:
        for pv_systems_xr, image_xr in Zipper(self.source_datapipe, self.image_datapipe):
            if 0 < self.max_num_pv_systems <= len(pv_systems_xr.pv_system_id.values):
                subset_of_pv_system_ids = self.rng.choice(
                    pv_systems_xr.pv_system_id,
                    size=self.max_num_pv_systems,
                    replace=False,
                )
                pv_systems_xr = pv_systems_xr.sel(pv_system_id=subset_of_pv_system_ids)

            if "geostationary" in self.x_dim:
                _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(
                    image_xr
                )
            # Create empty image to use for the PV Systems, assumes image has x and y coordinates
            pv_image = np.zeros(
                (
                    len(pv_systems_xr["time_utc"]),
                    len(image_xr[self.y_dim]),
                    len(image_xr[self.x_dim]),
                ),
                dtype=np.float32,
            )
            pv_position_dict = defaultdict(list)
            for i, pv_system_id in enumerate(pv_systems_xr["pv_system_id"]):
                try:
                    # went for isel incase there is a duplicated pv_system_id
                    pv_system = pv_systems_xr.isel(pv_system_id=i)
                except Exception as e:
                    logger.warning(
                        f"Could not select {pv_system_id} " f"from {pv_systems_xr.pv_system_id}"
                    )
                    raise e
                if "geostationary" in self.x_dim:
                    pv_x, pv_y = _osgb_to_geostationary(
                        xx=pv_system["x_osgb"].values, yy=pv_system["y_osgb"].values
                    )
                else:
                    pv_x = pv_system["x_osgb"]
                    pv_y = pv_system["y_osgb"]
                # Quick check as search sorted doesn't give an error if it is not in the range
                if pv_x < image_xr[self.x_dim][0].values or pv_x > image_xr[self.x_dim][-1].values:
                    continue
                # Y Coordinates are in reverse for satellite data
                if pv_y > image_xr[self.y_dim][0].values or pv_y < image_xr[self.y_dim][-1].values:
                    continue
                if "geostationary" in self.x_dim:
                    x_idx = np.searchsorted(image_xr[self.x_dim].values, pv_x) - 1
                    # y_geostationary is in descending order:
                    y_idx = len(image_xr[self.y_dim]) - (
                        np.searchsorted(image_xr[self.y_dim].values[::-1], pv_y) - 1
                    )
                else:
                    x_idx = np.searchsorted(pv_x, image_xr[self.x_dim])
                    y_idx = np.searchsorted(pv_y, image_xr[self.y_dim])
                if x_idx == len(image_xr[self.x_dim]) or y_idx == len(image_xr[self.y_dim]):
                    continue
                # Add location to same one, so know if multiple overlap
                # Filter if normalizing by pvlib,
                # only include ones that have tilt and orientation as numbers
                if self.normalize_by_pvlib:
                    if not np.isfinite(pv_system.orientation.values) and not np.isfinite(
                        pv_system.tilt.values
                    ):
                        continue
                pv_position_dict[(y_idx, x_idx)].append(pv_system)
            for location, system_list in pv_position_dict.items():
                y_idx, x_idx = location[0], location[1]
                if 0 < self.take_n_pv_values_per_pixel < len(system_list):
                    system_numbers = self.rng.choice(
                        list(range(len(system_list))),
                        size=self.take_n_pv_values_per_pixel,
                        replace=False,
                    )
                    system_list = [system_list[idx] for idx in system_numbers]
                avg_generation = np.zeros_like(system_list[0].values)
                for pv_system in system_list:
                    if self.normalize_by_pvlib:
                        pv_system = _normalize_by_pvlib(pv_system)
                    avg_generation += np.nan_to_num(pv_system.values)
                avg_generation /= len(system_list)
                pv_image[:, y_idx, x_idx] = avg_generation

            if self.normalize:
                if np.nanmax(pv_image) > 0:
                    pv_image /= np.nanmax(pv_image)
            pv_image = np.nan_to_num(pv_image)

            # Should return Xarray as in Xarray transforms
            # Same coordinates as the image xarray, so can take that
            pv_image = _create_data_array_from_image(pv_image, pv_systems_xr, image_xr)
            if self.always_return_first:
                while True:
                    yield pv_image
            yield pv_image


def _create_data_array_from_image(
    pv_image: np.ndarray,
    pv_systems_xr: Union[xr.Dataset, xr.DataArray],
    image_xr: Union[xr.Dataset, xr.DataArray],
):
    data_array = xr.DataArray(
        data=pv_image,
        coords=(
            ("time_utc", pv_systems_xr.time_utc.values),
            ("y_geostationary", image_xr.y_geostationary.values),
            ("x_geostationary", image_xr.x_geostationary.values),
        ),
        name="pv_image",
    ).astype(np.float32)
    data_array.attrs = image_xr.attrs
    return data_array


def _get_idx_of_pixel_closest_to_poi_geostationary(
    xr_data: xr.DataArray,
    center_osgb: Location,
    x_dim_name="x_geostationary",
    y_dim_name="y_geostationary",
) -> Location:
    """
    Return x and y index location of pixel at center of region of interest.

    Args:
        xr_data: Xarray dataset
        center_osgb: Center in OSGB coordinates
        x_dim_name: X dimension name
        y_dim_name: Y dimension name

    Returns:
        Location for the center pixel in geostationary coordinates
    """
    _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
    center_geostationary_tuple = _osgb_to_geostationary(xx=center_osgb.x, yy=center_osgb.y)
    center_geostationary = Location(
        x=center_geostationary_tuple[0], y=center_geostationary_tuple[1]
    )

    # Get the index into x and y nearest to x_center_geostationary and y_center_geostationary:
    x_index_at_center = np.searchsorted(xr_data[x_dim_name].values, center_geostationary.x) - 1
    # y_geostationary is in descending order:
    y_index_at_center = len(xr_data[y_dim_name]) - (
        np.searchsorted(xr_data[y_dim_name].values[::-1], center_geostationary.y) - 1
    )
    return Location(x=x_index_at_center, y=y_index_at_center)


def _normalize_by_pvlib(pv_system):
    """
    Normalize the output by pv_libs poa_global

    Args:
        pv_system: PV System in Xarray DataArray

    Returns:
        PV System in xarray DataArray, but normalized values
    """
    # TODO Add elevation
    pvlib_loc = pvlib.location.Location(
        latitude=pv_system.latitude.values, longitude=pv_system.longitude.values
    )
    times = pd.DatetimeIndex(pv_system.time_utc.values)
    solar_position = pvlib_loc.get_solarposition(times=times)
    clear_sky = pvlib_loc.get_clearsky(times)
    total_irradiance = pvlib.irradiance.get_total_irradiance(
        pv_system.tilt.values,
        pv_system.orientation.values,
        solar_zenith=solar_position["zenith"],
        solar_azimuth=solar_position["azimuth"],
        dni=clear_sky["dni"],
        dhi=clear_sky["dhi"],
        ghi=clear_sky["ghi"],
    )
    # Guess want fraction of total irradiance on panel, to get fraction to do with capacity
    fraction_clear_sky = total_irradiance["poa_global"] / (
        clear_sky["dni"] + clear_sky["dhi"] + clear_sky["ghi"]
    )
    pv_system /= pv_system.capacity_watt_power
    pv_system *= fraction_clear_sky
    return pv_system



"""Remove PV data """

import logging
from datetime import timedelta

from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("pv_remove_zero_data")
class PVPowerRemoveZeroDataIterDataPipe(IterDataPipe):
    """Compute rolling mean of PV power."""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        window: timedelta,
    ):
        """
        Set zero pv data to nans

        Args:
            source_datapipe: iter data pipe
            window: the time length of the window to check
        """

        self.source_datapipe = source_datapipe
        self.window = window

    def __iter__(self):
        for xr_data in self.source_datapipe:
            logger.debug(f"Reducing Date if window ({self.window}) of zeros")

            # get rolling window values
            window_length = int(self.window / xr_data.sample_period_duration)
            resampled = xr_data.rolling(dim={"time_utc": window_length}).max()

            xr_data = xr_data.where(resampled > 0.0)
            yield xr_data



"""Open GFS Forecast data"""
import logging
from pathlib import Path
from typing import Union

import pandas as pd
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

_log = logging.getLogger(__name__)


@functional_datapipe("open_gfs")
class OpenGFSForecastIterDataPipe(IterDataPipe):
    """Open GFS Forecast data"""

    def __init__(self, zarr_path: Union[Path, str]):
        """
        Open GFS Forecast data

        Args:
            zarr_path: Path or wildcard path to GFS Zarrs
        """
        self.zarr_path = zarr_path

    def __iter__(self):
        _log.debug("Opening NWP data: %s", self.zarr_path)
        gfs = open_gfs(self.zarr_path)

        while True:
            yield gfs


def open_gfs(zarr_path: Union[Path, str]) -> xr.Dataset:
    """
    Opens GFS dataset

    Args:
        zarr_path: Path to Zarr(s) to open

    Returns:
        Xarray dataset of GFS Forecasts
    """

    _log.info("Loading NWP GFS data")

    if "*" in zarr_path:
        nwp = xr.open_mfdataset(zarr_path, engine="zarr", combine="time", chunks="auto")
    else:
        nwp = xr.load_dataset(zarr_path, engine="zarr", mode="r", chunks="auto")

    variables = list(nwp.keys())

    nwp = xr.concat([nwp[v] for v in variables], "channel")
    nwp = nwp.assign_coords(channel=variables)

    nwp = nwp.transpose("time", "step", "channel", "latitude", "longitude")
    nwp = nwp.rename({"time": "init_time_utc"})
    nwp = nwp.transpose("init_time_utc", "step", "channel", "latitude", "longitude")
    if "valid_time" in nwp.coords.keys():
        nwp = nwp.drop("valid_time")

    # nwp = nwp.resample(init_time_utc="60T").pad()
    _log.debug("Interpolating hour 0 to NWP data")
    nwp_step0 = nwp.interp(step=[pd.Timedelta(hours=0)])
    nwp = xr.concat([nwp_step0, nwp], dim="step")
    nwp = nwp.resample(init_time_utc="60T").pad()
    nwp = nwp.resample(step="60T").pad()

    _log.debug(nwp)

    return nwp



"""Datapipes to add Sun position to NumpyBatch"""

import logging

import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("change_float32")
class ChangeFloat32IterDataPipe(IterDataPipe):
    """Change to float 64 to 32s"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Change float 64s to float 32s

        Args:
            source_datapipe: Datapipe of NumpyBatch
        """
        self.source_datapipe = source_datapipe

    def __iter__(self):
        for np_batch in self.source_datapipe:
            logger.debug("Changing arrays to float32s")

            for key in np_batch.keys():
                if isinstance(np_batch[key], np.ndarray):
                    if np_batch[key].dtype == np.float64:
                        np_batch[key] = np_batch[key].astype("float32")

            yield np_batch



"""Satellite loader"""
import logging
import subprocess
from datetime import datetime, timedelta
from pathlib import Path
from typing import Union

import dask
import fsspec
import pandas as pd
import xarray as xr
from ocf_blosc2 import Blosc2  # noqa: F401
from pathy import Pathy
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

_log = logging.getLogger(__name__)


def _get_single_sat_data(zarr_path: Union[Path, str]) -> xr.DataArray:
    """Helper function to open a zarr from either local or GCP path.

    The local or GCP path may contain wildcard matching (*)

    Args:
        zarr_path: Path to zarr file
    """

    # These kwargs are used if zarr path contains "*"
    openmf_kwargs = dict(
        engine="zarr",
        concat_dim="time",
        combine="nested",
        chunks="auto",
        join="override",
    )

    # Need to generate list of files if using GCP bucket storage
    if "gs://" in str(zarr_path) and "*" in str(zarr_path):
        result_string = subprocess.run(
            f"gsutil ls -d {zarr_path}".split(" "), stdout=subprocess.PIPE
        ).stdout.decode("utf-8")
        files = result_string.splitlines()

        dataset = xr.open_mfdataset(files, **openmf_kwargs)

    elif "*" in str(zarr_path):  # Multi-file dataset
        dataset = xr.open_mfdataset(zarr_path, **openmf_kwargs)
    else:
        dataset = xr.open_dataset(zarr_path, engine="zarr", chunks="auto")
    dataset = dataset.drop_duplicates("time").sortby("time")

    return dataset


def open_sat_data(
    zarr_path: Union[Path, str, list[Path], list[str]], use_15_minute_data_if_needed: bool = False
) -> xr.DataArray:
    """Lazily opens the Zarr store.

    Args:
      zarr_path: Cloud URL or local path pattern, or list of these. If GCS URL, it must start with
          'gs://'.
      use_15_minute_data_if_needed: use_15_minute_data_if_needed: Option to use the 15 minute data
        if the 5 minute data is not available
        This is done by checking to see if the last timestamp is within an hour from now

    Example:
        With wild cards and GCS path:
        ```
        zarr_paths = [
            "gs://bucket/2020_nonhrv_split_*.zarr",
            "gs://bucket/2019_nonhrv_split_*.zarr",
        ]
        ds = open_sat_data(zarr_paths)
        ```
        Without wild cards and with local path:
        ```
        zarr_paths = [
            "/data/2020_nonhrv.zarr",
            "/data/2019_nonhrv.zarr",
        ]
        ds = open_sat_data(zarr_paths)
        ```
    """
    _log.info("Opening satellite data: %s", zarr_path)

    # Silence the warning about large chunks.
    # Alternatively, we could set this to True, but that slows down loading a Satellite batch
    # from 8 seconds to 50 seconds!
    dask.config.set({"array.slicing.split_large_chunks": False})

    # dont load 15 minute data by default
    use_15_minute_data = False

    if isinstance(zarr_path, (list, tuple)):
        dataset = xr.combine_nested(
            [_get_single_sat_data(path) for path in zarr_path],
            concat_dim="time",
            combine_attrs="override",
            join="override",
        )
    else:
        # check the file exists
        dataset, use_15_minute_data = load_and_check_satellite_data(zarr_path)

    if use_15_minute_data_if_needed and use_15_minute_data:
        zarr_path_15_minutes = str(zarr_path).replace(".zarr", "_15.zarr")

        _log.info(f"Now going to load {zarr_path_15_minutes} and resample")
        dataset = _get_single_sat_data(zarr_path_15_minutes)

        dataset = dataset.load()
        _log.debug("Resampling 15 minute data to 5 mins")
        dataset = dataset.resample(time="5T").interpolate("linear")

    # Remove data coordinate dimensions if they exist
    if "x_geostationary_coordinates" in dataset:
        del dataset["x_geostationary_coordinates"]
        del dataset["y_geostationary_coordinates"]

    # Rename
    # These renamings will no longer be necessary when the Zarr uses the 'correct' names,
    # see https://github.com/openclimatefix/Satip/issues/66
    if "variable" in dataset:
        dataset = dataset.rename({"variable": "channel"})
    if "channels" in dataset:
        dataset = dataset.rename({"channels": "channel"})
    elif "channel" not in dataset:
        # This is HRV version 3, which doesn't have a channels dim.  So add one.
        dataset = dataset.expand_dims(dim={"channel": ["HRV"]}, axis=1)

    # Rename coords to be more explicit about exactly what some coordinates hold:
    # Note that `rename` renames *both* the coordinates and dimensions, and keeps
    # the connection between the dims and coordinates, so we don't have to manually
    # use `data_array.set_index()`.
    dataset = dataset.rename(
        {
            "time": "time_utc",
        }
    )
    if "y" in dataset.coords.keys():
        dataset = dataset.rename(
            {
                "y": "y_geostationary",
            }
        )

    if "x" in dataset.coords.keys():
        dataset = dataset.rename(
            {
                "x": "x_geostationary",
            }
        )

    # Flip coordinates to top-left first
    if dataset.y_geostationary[0] < dataset.y_geostationary[-1]:
        dataset = dataset.reindex(y_geostationary=dataset.y_geostationary[::-1])
    if dataset.x_geostationary[0] > dataset.x_geostationary[-1]:
        dataset = dataset.reindex(x_geostationary=dataset.x_geostationary[::-1])

    data_array = dataset["data"]
    del dataset

    # Ensure the y and x coords are in the right order (top-left first):
    assert data_array.y_geostationary[0] > data_array.y_geostationary[-1]
    assert data_array.x_geostationary[0] < data_array.x_geostationary[-1]
    if "y_osgb" in data_array.dims:
        assert data_array.y_osgb[0, 0] > data_array.y_osgb[-1, 0]
        assert data_array.x_osgb[0, 0] < data_array.x_osgb[0, -1]

    # Sanity checks!
    data_array = data_array.transpose("time_utc", "channel", "y_geostationary", "x_geostationary")
    assert data_array.dims == ("time_utc", "channel", "y_geostationary", "x_geostationary")
    datetime_index = pd.DatetimeIndex(data_array.time_utc)
    assert datetime_index.is_unique
    assert datetime_index.is_monotonic_increasing
    # Satellite datetimes can sometimes be 04, 09, minutes past the hour, or other slight offsets.
    # These slight offsets will break downstream code, which expects satellite data to be at
    # exactly 5 minutes past the hour.
    assert (datetime_index == datetime_index.round("5T")).all()

    _log.info("Opened satellite data")

    return data_array


def load_and_check_satellite_data(zarr_path) -> [xr.Dataset, bool]:
    """
    Load the satellite data,

    1. check if file path exists
    2. check dataset has been updated in the last hour
    If 1. or 2. are true, then return True for use_15_minute_data

    Args:
        zarr_path: the zarr path to load

    Returns:
        dataset (if loaded),
        use_15_minute_data, indicating if the 15 minute data should be loaded
    """
    filesystem = fsspec.open(Pathy.fluid(zarr_path)).fs
    if filesystem.exists(zarr_path):
        dataset = _get_single_sat_data(zarr_path)

        use_15_minute_data = check_last_timestamp(dataset)

    else:
        _log.info(f"File does not exist {zarr_path}. Will try to load 15 minute data")
        use_15_minute_data = True
        dataset = None
    return dataset, use_15_minute_data


def check_last_timestamp(dataset: xr.Dataset, timedelta_hours: float = 1) -> bool:
    """
    Check the last timestamp of the dataset to see if it is more than 1 hour ago

    Args:
        dataset: dataset with time dimension
        timedelta_hours: the timedelta to check from now

    Returns: bool
    """
    latest_time = pd.to_datetime(dataset.time[-1].values)
    now = datetime.utcnow()
    if latest_time < now - timedelta(hours=timedelta_hours):
        _log.info(
            f"last datestamp is {latest_time}, which is more than "
            f"{timedelta_hours} hour ago from {now} "
            f"Will try to load 15 minute data"
        )
        return True
    else:
        return False


@functional_datapipe("open_satellite")
class OpenSatelliteIterDataPipe(IterDataPipe):
    """Open Satellite Zarr"""

    def __init__(self, zarr_path: Union[Path, str], use_15_minute_data_if_needed: bool = False):
        """
        Opens the satellite Zarr

        Args:
            zarr_path: path to the zarr file
            use_15_minute_data_if_needed: Option to use the 15 minute data if the
                5 minute data is not available
        """
        self.zarr_path = zarr_path
        self.use_15_minute_data_if_needed = use_15_minute_data_if_needed
        super().__init__()

    def __iter__(self) -> xr.DataArray:
        """Open the Zarr file"""
        data: xr.DataArray = open_sat_data(
            zarr_path=self.zarr_path, use_15_minute_data_if_needed=self.use_15_minute_data_if_needed
        )
        while True:
            yield data



"""Convert point PV site metadata to image output"""
import logging
from typing import Union

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils import Zipper
from ocf_datapipes.utils.consts import Location
from ocf_datapipes.utils.geospatial import load_geostationary_area_definition_and_transform_osgb

logger = logging.getLogger(__name__)

MAX_TILT = 90.0
MAX_ORIENTATION = 360.0


@functional_datapipe("create_pv_metadata_image")
class CreatePVMetadataImageIterDataPipe(IterDataPipe):
    """Create PV Metadata image from individual sites"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        image_datapipe: IterDataPipe,
        image_dim: str = "geostationary",
        max_num_pv_systems: int = -1,
        always_return_first: bool = False,
        normalize: bool = False,
        seed: int = None,
    ):
        """
        Creates a 2D data cube of PV Metadata output image

        The metadata is orientation + tilt

        Args:
            source_datapipe: Source datapipe of PV data
            image_datapipe: Datapipe emitting images to get the shape from, with coordinates
            image_dim: Dimension name for the x and y dimensions
            max_num_pv_systems: Max number of PV systems to use
            normalize: Whether to normalize tilt and orientation to between 0 and 1
            always_return_first: Always return the first image data cube, to save computation
                Only use for if making the image at the beginning of the stack
            seed: Random seed to use if using max_num_pv_systems
        """
        self.source_datapipe = source_datapipe
        self.image_datapipe = image_datapipe
        self.x_dim = "x_" + image_dim
        self.y_dim = "y_" + image_dim
        self.max_num_pv_systems = max_num_pv_systems
        self.rng = np.random.default_rng(seed=seed)
        self.always_return_first = always_return_first
        self.normalize = normalize

    def __iter__(self) -> xr.DataArray:
        for pv_systems_xr, image_xr in Zipper(self.source_datapipe, self.image_datapipe):
            if 0 < self.max_num_pv_systems <= len(pv_systems_xr.pv_system_id.values):
                subset_of_pv_system_ids = self.rng.choice(
                    pv_systems_xr.pv_system_id,
                    size=self.max_num_pv_systems,
                    replace=False,
                )
                pv_systems_xr = pv_systems_xr.sel(pv_system_id=subset_of_pv_system_ids)

            if "geostationary" in self.x_dim:
                _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(
                    image_xr
                )
            # Create empty image to use for the PV Systems, assumes image has x and y coordinates
            pv_image = np.zeros(
                (
                    2,
                    len(image_xr[self.y_dim]),
                    len(image_xr[self.x_dim]),
                ),
                dtype=np.float32,
            )
            for i, pv_system_id in enumerate(pv_systems_xr["pv_system_id"]):
                try:
                    # went for isel incase there is a duplicated pv_system_id
                    pv_system = pv_systems_xr.isel(pv_system_id=i)
                except Exception as e:
                    logger.warning(
                        f"Could not select {pv_system_id} " f"from {pv_systems_xr.pv_system_id}"
                    )
                    raise e
                if "geostationary" in self.x_dim:
                    pv_x, pv_y = _osgb_to_geostationary(
                        xx=pv_system["x_osgb"].values, yy=pv_system["y_osgb"].values
                    )
                else:
                    pv_x = pv_system["x_osgb"]
                    pv_y = pv_system["y_osgb"]
                # Quick check as search sorted doesn't give an error if it is not in the range
                if pv_x < image_xr[self.x_dim][0].values or pv_x > image_xr[self.x_dim][-1].values:
                    continue
                # Y Coordinates are in reverse for satellite data
                if pv_y > image_xr[self.y_dim][0].values or pv_y < image_xr[self.y_dim][-1].values:
                    continue
                if "geostationary" in self.x_dim:
                    x_idx = np.searchsorted(image_xr[self.x_dim].values, pv_x) - 1
                    # y_geostationary is in descending order:
                    y_idx = len(image_xr[self.y_dim]) - (
                        np.searchsorted(image_xr[self.y_dim].values[::-1], pv_y) - 1
                    )
                else:
                    x_idx = np.searchsorted(pv_x, image_xr[self.x_dim])
                    y_idx = np.searchsorted(pv_y, image_xr[self.y_dim])
                if x_idx == len(image_xr[self.x_dim]) or y_idx == len(image_xr[self.y_dim]):
                    continue
                # Now go by the timestep to create cube of PV data
                pv_image[:, y_idx, x_idx] = np.array(pv_system["tilt"], pv_system["orientation"])

            if self.normalize:
                pv_image[0, :, :] = pv_image[0, :, :] / MAX_TILT
                pv_image[1, :, :] = pv_image[1, :, :] / MAX_ORIENTATION
            pv_image = np.nan_to_num(pv_image)

            # Should return Xarray as in Xarray transforms
            # Same coordinates as the image xarray, so can take that
            pv_image = _create_data_array_from_image(pv_image, pv_systems_xr, image_xr)
            if self.always_return_first:
                while True:
                    yield pv_image
            yield pv_image


def _create_data_array_from_image(
    pv_image: np.ndarray,
    pv_systems_xr: Union[xr.Dataset, xr.DataArray],
    image_xr: Union[xr.Dataset, xr.DataArray],
):
    data_array = xr.DataArray(
        data=pv_image,
        coords=(
            ("time_utc", ["tilt", "orientation"]),
            ("y_geostationary", image_xr.y_geostationary.values),
            ("x_geostationary", image_xr.x_geostationary.values),
        ),
        name="pv_meta_image",
    ).astype(np.float32)
    data_array.attrs = image_xr.attrs
    return data_array


def _get_idx_of_pixel_closest_to_poi_geostationary(
    xr_data: xr.DataArray,
    center_osgb: Location,
    x_dim_name="x_geostationary",
    y_dim_name="y_geostationary",
) -> Location:
    """
    Return x and y index location of pixel at center of region of interest.

    Args:
        xr_data: Xarray dataset
        center_osgb: Center in OSGB coordinates
        x_dim_name: X dimension name
        y_dim_name: Y dimension name

    Returns:
        Location for the center pixel in geostationary coordinates
    """
    _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
    center_geostationary_tuple = _osgb_to_geostationary(xx=center_osgb.x, yy=center_osgb.y)
    center_geostationary = Location(
        x=center_geostationary_tuple[0], y=center_geostationary_tuple[1]
    )

    # Get the index into x and y nearest to x_center_geostationary and y_center_geostationary:
    x_index_at_center = np.searchsorted(xr_data[x_dim_name].values, center_geostationary.x) - 1
    # y_geostationary is in descending order:
    y_index_at_center = len(xr_data[y_dim_name]) - (
        np.searchsorted(xr_data[y_dim_name].values[::-1], center_geostationary.y) - 1
    )
    return Location(x=x_index_at_center, y=y_index_at_center)



from ocf_datapipes.transform.xarray import Downsample


def test_nwp_downsample(nwp_datapipe):
    nwp_datapipe = Downsample(nwp_datapipe, y_coarsen=16, x_coarsen=16)
    data = next(iter(nwp_datapipe))
    # Downsample by 16 from 704x548
    assert data.shape[-1] == 34
    assert data.shape[-2] == 44


def test_sat_downsample(sat_datapipe):
    sat_datapipe = Downsample(
        sat_datapipe,
        y_coarsen=16,
        x_coarsen=16,
        y_dim_name="y_geostationary",
        x_dim_name="x_geostationary",
    )
    data = next(iter(sat_datapipe))
    assert data.shape[-1] == 38
    assert data.shape[-2] == 18


def test_topo_downsample(topo_datapipe):
    topo_datapipe = Downsample(topo_datapipe, y_coarsen=16, x_coarsen=16)
    data = next(iter(topo_datapipe))
    assert data.shape == (176, 272)



"""Select the history for the live data"""
import logging
from datetime import timedelta
from typing import Union

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils import Zipper

logger = logging.getLogger(__name__)


@functional_datapipe("select_live_time_slice")
class SelectLiveTimeSliceIterDataPipe(IterDataPipe):
    """Select the history for the live data"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        t0_datapipe: IterDataPipe,
        history_duration: timedelta,
        dim_name: str = "time_utc",
    ):
        """
        Select the history for the live time slice

        Args:
            source_datapipe: Datapipe emitting Xarray object
            t0_datapipe: Datapipe emitting t0 timestamps
            history_duration: Amount of time for the history
            dim_name: Time dimension name
        """
        self.source_datapipe = source_datapipe
        self.t0_datapipe = t0_datapipe
        self.history_duration = np.timedelta64(history_duration)
        self.dim_name = dim_name

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Select the recent live data"""
        for xr_data, t0 in Zipper(self.source_datapipe, self.t0_datapipe):
            logger.debug(f"Selecting time slice {t0} on dim {self.dim_name}")

            xr_data = xr_data.sel({self.dim_name: slice(t0 - self.history_duration, t0)})

            logger.debug(f"Took slice of length {len(getattr(xr_data,self.dim_name))}")

            yield xr_data



from ocf_datapipes.config.model import PV, PVFiles
from ocf_datapipes.load import OpenPVFromNetCDF


def test_remove_nans_all_pv_systems(pv_parquet_file):
    pv = PV()
    pv_file = PVFiles(
        pv_filename="tests/data/pv/passiv/test.nc",
        pv_metadata_filename="tests/data/pv/passiv/UK_PV_metadata.csv",
        label="solar_sheffield_passiv",
    )
    pv.pv_files_groups = [pv_file]

    pv_datapipe = OpenPVFromNetCDF(pv=pv)

    pv_datapipe = pv_datapipe.remove_nans()
    pv_datapipe = pv_datapipe.check_nans()

    data = next(iter(pv_datapipe))
    assert data is not None



"""Set system IDs to all 1"""
import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch


@functional_datapipe("set_system_ids_to_one")
class SetSystemIDsToOneIterDataPipe(IterDataPipe):
    """Set system IDs to all 1"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Set all system IDs to 1 as a temp measure for PP production

        Args:
            source_datapipe: Datapipe of NumpyBatch objects
        """
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NumpyBatch:
        """Set system IDs to all 1"""
        for np_batch in self.source_datapipe:
            # np_batch[BatchKey.gsp_id] = np.ones_like(np_batch[BatchKey.gsp_id])
            np_batch[BatchKey.pv_id] = np.ones_like(np_batch[BatchKey.pv_id])
            yield np_batch



from datetime import timedelta
from torchdata.datapipes.iter import IterableWrapper
import pandas as pd
import numpy as np

from ocf_datapipes.transform.xarray import SelectDropoutTime, ApplyDropoutTime


def test_select_dropout_time(sat_datapipe):
    data = next(iter(sat_datapipe))

    t0_times = pd.to_datetime(data.time_utc.values)

    t0_datapipe = IterableWrapper(t0_times)

    # All times delayed by 5 minutes
    dropout_time_datapipe = SelectDropoutTime(
        source_datapipe=t0_datapipe,
        dropout_timedeltas=[timedelta(minutes=-5)],
        dropout_frac=1,
    )

    dropout_times = [*dropout_time_datapipe]

    assert (np.array(dropout_times) == t0_times - timedelta(minutes=5)).all()


def test_apply_dropout_time(sat_datapipe):
    data = next(iter(sat_datapipe)).isel(x_geostationary=0, y_geostationary=0).compute()

    times = pd.to_datetime(data.time_utc.values[3:])

    dropout_time_datapipe = IterableWrapper(times - timedelta(minutes=5))

    sliced_sat_datapipe = IterableWrapper(
        [data.sel(time_utc=slice(t - timedelta(minutes=10), t)) for t in times]
    )

    dropout_out_sat = ApplyDropoutTime(
        source_datapipe=sliced_sat_datapipe,
        dropout_time_datapipe=dropout_time_datapipe,
    )

    # Only the last element of the slice should be nan
    for sat_im in dropout_out_sat:
        assert (np.isnan(sat_im.values).squeeze() == np.array([False, False, True])).all()



"""Select the history for the live data"""
import pandas as pd
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe


@functional_datapipe("select_live_t0_time")
class SelectLiveT0TimeIterDataPipe(IterDataPipe):
    """Select the history for the live data"""

    def __init__(self, source_datapipe: IterDataPipe, dim_name: str = "time_utc"):
        """
        Select history for the Xarray object

        Args:
            source_datapipe: Datapipe emitting Xarray objects
            dim_name: The time dimension name to use
        """
        self.source_datapipe = source_datapipe
        self.dim_name = dim_name

    def __iter__(self) -> pd.Timestamp:
        """Get the latest timestamp and return it"""
        for xr_data in self.source_datapipe:
            # Get most recent time in data
            # Select the history that goes back that far
            latest_time_idx = pd.DatetimeIndex(xr_data[self.dim_name].values).get_indexer(
                [pd.Timestamp.now(tz=None)], method="pad"
            )[0]
            latest_time = xr_data[self.dim_name].values[latest_time_idx]
            yield latest_time



"""Datapipes for batching together data"""
from .merge_numpy_examples_to_batch import (
    MergeNumpyExamplesToBatchIterDataPipe as MergeNumpyExamplesToBatch,
)
from .merge_numpy_modalities import MergeNumpyModalitiesIterDataPipe as MergeNumpyModalities



from datetime import timedelta
from torchdata.datapipes.iter import IterableWrapper
import pandas as pd

from ocf_datapipes.transform.xarray import ConvertToNWPTargetTimeWithDropout


def test_convert_to_nwp_target_time_with_dropout(nwp_datapipe):
    data = next(iter(nwp_datapipe))

    t0 = pd.Timestamp(data.init_time_utc.values[3])
    times = [
        t0 - timedelta(minutes=60),
        t0,
        t0 + timedelta(minutes=30),
        t0 + timedelta(minutes=60),
        t0 + timedelta(minutes=120),
    ]

    t0_datapipe = IterableWrapper(times)

    dropout_datapipe = ConvertToNWPTargetTimeWithDropout(
        nwp_datapipe,
        t0_datapipe,
        sample_period_duration=timedelta(minutes=60),
        history_duration=timedelta(minutes=60),
        forecast_duration=timedelta(minutes=60),
        dropout_timedeltas=[timedelta(minutes=-5)],
        dropout_frac=1,
    )

    # The init-times should be the last 3-hour mutliple before 5-minutes before each time t
    for t, ds in zip(times, dropout_datapipe):
        assert (
            ds.init_time_utc.values == (t - timedelta(minutes=5)).floor(timedelta(hours=3))
        ).all()

    times = [
        t0,
        t0 + timedelta(minutes=30),
        t0 + timedelta(minutes=120),
    ]

    t0_datapipe = IterableWrapper(times)

    # Dropout with one hour delay
    dropout_datapipe = ConvertToNWPTargetTimeWithDropout(
        nwp_datapipe,
        t0_datapipe,
        sample_period_duration=timedelta(minutes=60),
        history_duration=timedelta(minutes=60),
        forecast_duration=timedelta(minutes=0),
        dropout_timedeltas=[timedelta(minutes=-55)],
        dropout_frac=1,
    )

    # The init times should be the last 3-hour mutliple before an hour before each time t
    for t, ds in zip(times, dropout_datapipe):
        assert (
            ds.init_time_utc.values == (t - timedelta(minutes=60)).floor(timedelta(hours=3))
        ).all()



"""Wrapper for Power Perceiver Production Data Pipeline"""
import logging
from datetime import timedelta
from pathlib import Path
from typing import Optional, Union

import xarray
from torchdata.datapipes.iter import IterDataPipe

import ocf_datapipes  # noqa
from ocf_datapipes.batch import MergeNumpyModalities
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenConfiguration, OpenPVFromDB, OpenPVFromNetCDF

logger = logging.getLogger(__name__)
xarray.set_options(keep_attrs=True)

# default is set to 1000
BUFFERSIZE = 1000


def simple_pv_datapipe(
    configuration_filename: Union[Path, str], tag: Optional[str] = "train"
) -> IterDataPipe:
    """
    Create the Power Perceiver production pipeline using a configuration

    Args:
        configuration_filename: Name of the configuration
        tag: Tag for train or test

    Returns:
        DataPipe ready to be put in a Dataloader for production
    """
    ####################################
    #
    # main data pipe for loading a simple site level forecast
    #
    #####################################
    # load configuration
    config_datapipe = OpenConfiguration(configuration_filename)
    configuration: Configuration = next(iter(config_datapipe))

    logger.debug("Opening Datasets")
    if configuration.input_data.pv.is_live:
        pv_datapipe = OpenPVFromDB(pv_config=configuration.input_data.pv)
    else:
        pv_datapipe = OpenPVFromNetCDF(pv=configuration.input_data.pv)
    pv_datapipe, pv_location_datapipe = pv_datapipe.fork(2, buffer_size=BUFFERSIZE)

    logger.debug("Add t0 idx")
    pv_datapipe = pv_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
    )

    logger.debug("Getting locations")
    # might have to fork this if we add NWPs
    location_datapipe1 = pv_location_datapipe.location_picker(
        x_dim_name="longitude",
        y_dim_name="latitude",
    )
    logger.debug("Got locations")

    logger.debug("Making PV space slice")
    pv_datapipe, pv_t0_datapipe, pv_time_periods_datapipe = (
        pv_datapipe.normalize(normalize_fn=lambda x: x / x.capacity_watt_power)
        .add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(
                minutes=configuration.input_data.pv.time_resolution_minutes
            ),
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        )
        .select_id(location_datapipe=location_datapipe1, data_source_name="pv")
        # .select_spatial_slice_meters(
        #     location_datapipe=location_datapipe1,
        #     roi_width_meters=configuration.input_data.pv.pv_image_size_meters_width,
        #     roi_height_meters=configuration.input_data.pv.pv_image_size_meters_height,
        #     x_dim_name="longitude", y_dim_name="latitude"
        # )
        .ensure_n_pv_systems_per_example(n_pv_systems_per_example=1)
        .remove_nans()
        .fork(3, buffer_size=BUFFERSIZE)
    )

    # get contiguous time periods
    pv_time_periods_datapipe = pv_time_periods_datapipe.get_contiguous_time_periods(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
    )

    # select time periods
    pv_t0_datapipe = pv_t0_datapipe.select_time_periods(time_periods=pv_time_periods_datapipe)
    pv_t0_datapipe = pv_t0_datapipe.select_t0_time()

    # take time slices
    pv_datapipe = (
        pv_datapipe.select_time_slice(
            t0_datapipe=pv_t0_datapipe,
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
            forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
            sample_period_duration=timedelta(
                minutes=configuration.input_data.pv.time_resolution_minutes
            ),
        )
        .convert_pv_to_numpy_batch()
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )

    ####################################
    #
    # Join data pipes together, and get extra details
    # TODO add simple NWP data
    #
    #####################################
    logger.debug("Combine all the data sources")
    combined_datapipe = (
        MergeNumpyModalities([pv_datapipe])
        # .align_gsp_to_5_min(batch_key_for_5_min_datetimes=BatchKey.pv_time_utc)
        .encode_space_time().add_sun_position(modality_name="pv")
    )

    # combined_datapipe = combined_datapipe.add_length(
    #     configuration=configuration, train_validation_test=tag
    # )

    return combined_datapipe



"""Pick locations from a dataset"""
import logging
from typing import Optional

import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import Location

logger = logging.getLogger(__name__)


@functional_datapipe("location_picker")
class LocationPickerIterDataPipe(IterDataPipe):
    """Picks locations from a dataset and returns them"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        return_all_locations: bool = False,
        x_dim_name: Optional[str] = "x_osgb",
        y_dim_name: Optional[str] = "y_osgb",
    ):
        """
        Picks locations from a dataset and returns them

        Args:
            source_datapipe: Datapipe emitting Xarray Dataset
            return_all_locations: Whether to return all locations,
                if True, also returns them in order
            x_dim_name: x dimension name, defaulted to 'x_osgb'
            y_dim_name: y dimension name, defaulted to 'y_osgb'
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.return_all_locations = return_all_locations
        self.x_dim_name = x_dim_name
        self.y_dim_name = y_dim_name

    def __iter__(self) -> Location:
        """Returns locations from the inputs datapipe"""
        for xr_dataset in self.source_datapipe:
            logger.debug("Getting locations")

            if self.return_all_locations:
                logger.debug("Going to return all locations")

                # Iterate through all locations in dataset
                for location_idx in range(len(xr_dataset[self.x_dim_name])):
                    location = Location(
                        x=xr_dataset[self.x_dim_name][location_idx].values,
                        y=xr_dataset[self.y_dim_name][location_idx].values,
                    )
                    if "pv_system_id" in xr_dataset.coords.keys():
                        location.id = int(xr_dataset["pv_system_id"][location_idx].values)
                    logger.debug(f"Got all location {location}")
                    yield location
            else:
                # Assumes all datasets have osgb coordinates for selecting locations
                # Pick 1 random location from the input dataset
                logger.debug("Selecting random idx")
                location_idx = np.random.randint(0, len(xr_dataset[self.x_dim_name]))
                logger.debug(f"{location_idx=}")
                location = Location(
                    x=xr_dataset[self.x_dim_name][location_idx].values,
                    y=xr_dataset[self.y_dim_name][location_idx].values,
                )
                if "pv_system_id" in xr_dataset.coords.keys():
                    location.id = int(xr_dataset["pv_system_id"][location_idx].values)
                    logger.debug(f"Have selected location.id {location.id}")
                logger.debug(f"{location=}")
                yield location



"""
This is a class function that assigns day night status
"""
import logging

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)

uk_daynight_dict = {
    1: [7, 17],
    2: [7, 18],
    3: [9, 19],
    4: [6, 20],
    5: [5, 21],
    6: [4, 22],
    7: [5, 22],
    8: [5, 21],
    9: [6, 20],
    10: [7, 21],
    11: [7, 17],
    12: [7, 17],
}
logger.info(
    f"The day and night standard hours are set by {'https://www.timeanddate.com/sun/uk/london'}"
)


@functional_datapipe("assign_daynight_status")
class AssignDayNightStatusIterDataPipe(IterDataPipe):
    """Adds a new dimension of day/night status"""

    def __init__(self, source_datapipe: IterDataPipe):
        """This method adds extra coordinate of day night status.

        Args:
            source_datapipe: Datapipe emiiting Xarray Dataset
        Result:
            <xarray.Dataset>
            Dimensions:   (datetime: 289)
            Coordinates:
            * datetime  (datetime) datetime64[ns]
            * status_day (datetime)
        """

        self.source_datapipe = source_datapipe

    def __iter__(self) -> xr.DataArray():
        """Returns an xarray dataset with extra dimesion"""

        # Reading the Xarray dataset
        for xr_dataset in self.source_datapipe:
            # Getting Month and Hour values from time_utc and stacking them
            date_month = xr_dataset.time_utc.dt.month.values
            date_hr = xr_dataset.time_utc.dt.hour.values
            month_hr_stack = np.stack((date_month, date_hr))

            # Getting the status of day/night for each timestamp in the dates'
            day_start = np.asarray([uk_daynight_dict[m][0] for m in month_hr_stack[0]])
            day_end = np.asarray([uk_daynight_dict[m][1] for m in month_hr_stack[0]])
            status_daynight = np.where(
                np.logical_and(month_hr_stack[1] >= day_start, month_hr_stack[1] < day_end),
                "day",
                "night",
            )

            # Assigning a new coordinates of 'status_daynight' in the DataArray
            xr_dataset = xr_dataset.assign_coords(status_daynight=(("time_utc"), status_daynight))
            yield xr_dataset



from datetime import timedelta

from ocf_datapipes.select import DropGSP, LocationPicker
from ocf_datapipes.transform.xarray import GetContiguousT0TimePeriods


def test_get_contiguous_time_periods(nwp_datapipe):
    nwp_datapipe = GetContiguousT0TimePeriods(
        nwp_datapipe,
        sample_period_duration=timedelta(hours=3),
        history_duration=timedelta(minutes=60),
        forecast_duration=timedelta(minutes=180),
        time_dim="init_time_utc",
    )

    batch = next(iter(nwp_datapipe))
    print(batch)



""" Util functions for PV data source"""
import logging
from typing import List, Optional

import numpy as np
import pandas as pd
import xarray as xr
from nowcasting_datamodel.models.pv import providers

logger = logging.getLogger(__name__)


def intersection_of_pv_system_ids(
    pv_metadata: pd.DataFrame, pv_power: pd.DataFrame
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Only pick PV systems for which we have metadata."""
    pv_system_ids = pv_metadata.index.intersection(pv_power.columns)
    pv_system_ids = np.sort(pv_system_ids)
    pv_power = pv_power[pv_system_ids]
    pv_metadata = pv_metadata.loc[pv_system_ids]
    return pv_metadata, pv_power


def put_pv_data_into_an_xr_dataarray(
    pv_power_watts: pd.DataFrame,
    y_osgb: pd.Series,
    x_osgb: pd.Series,
    capacity_watt_power: pd.Series,
    pv_system_row_number: pd.Series,
    longitude: Optional[pd.Series] = None,
    latitude: Optional[pd.Series] = None,
    tilt: Optional[pd.Series] = None,
    orientation: Optional[pd.Series] = None,
) -> xr.DataArray:
    """Convert to an xarray DataArray.

    Args:
        pv_power_watts: pd.DataFrame where the columns are PV systems (and the column names are
            ints), and the index is UTC datetime.
        x_osgb: The x location. Index = PV system ID ints.
        y_osgb: The y location. Index = PV system ID ints.
        capacity_watt_power: The max power output of each PV system in Watts.
         Index = PV system ID ints.
        pv_system_row_number: The integer position of the PV system in the metadata.
            Used to create the PV system ID embedding.
        longitude: longitude of the locations
        latitude: latitude of the locations
        tilt: Tilt of the panels
        orientation: Orientation of the panels
    """
    # Sanity check!
    pv_system_ids = pv_power_watts.columns
    for name, series in (
        ("x_osgb", x_osgb),
        ("y_osgb", y_osgb),
        ("capacity_watt_power", capacity_watt_power),
        ("pv_system_row_number", pv_system_row_number),
    ):
        logger.debug(f"Checking {name}")
        if not np.array_equal(series.index, pv_system_ids, equal_nan=True):
            logger.debug(f"Index of {name} does not equal {pv_system_ids}. Index is {series.index}")
            assert np.array_equal(series.index, pv_system_ids, equal_nan=True)

    data_array = xr.DataArray(
        data=pv_power_watts.values,
        coords=(
            ("time_utc", pv_power_watts.index.values),
            ("pv_system_id", pv_power_watts.columns),
        ),
        name="pv_power_watts",
    ).astype(np.float32)

    data_array = data_array.assign_coords(
        x_osgb=("pv_system_id", x_osgb),
        y_osgb=("pv_system_id", y_osgb),
        capacity_watt_power=("pv_system_id", capacity_watt_power),
        pv_system_row_number=("pv_system_id", pv_system_row_number),
    )
    # Sample period duration is required so PVDownsample transform knows by how much
    # to change the pv_t0_idx:
    if latitude is not None:
        data_array = data_array.assign_coords(
            latitude=("pv_system_id", latitude),
        )
    if longitude is not None:
        data_array = data_array.assign_coords(
            longitude=("pv_system_id", longitude),
        )
    if tilt is not None:
        data_array = data_array.assign_coords(
            tilt=("pv_system_id", tilt),
        )
    if orientation is not None:
        data_array = data_array.assign_coords(
            orientation=("pv_system_id", orientation),
        )

    assert len(pv_system_row_number) > 0

    return data_array


def encode_label(indexes: List[str], label: str) -> List[str]:
    """
    Encode the label to a list of indexes.

    The new encoding must be integers and unique.
    It would be useful if the indexes can read and deciphered by humans.
    This is done by times the original index by 10
    and adding 1 for passiv or 2 for other lables

    Args:
        indexes: list of indexes
        label: either 'solar_sheffield_passiv' or 'pvoutput.org'

    Returns: list of indexes encoded by label
    """
    assert label in providers
    # this encoding does work if the number of pv providers is more than 10
    assert len(providers) < 10

    label_index = providers.index(label)
    new_index = [int(int(col) * 10 + label_index) for col in indexes]

    return new_index



"""Selects time slice"""
import logging
from datetime import timedelta
from typing import Optional, Union

import numpy as np
import pandas as pd
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("select_time_slice")
class SelectTimeSliceIterDataPipe(IterDataPipe):
    """Selects time slice"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        t0_datapipe: IterDataPipe,
        sample_period_duration: timedelta,
        history_duration: Optional[timedelta] = None,
        forecast_duration: Optional[timedelta] = None,
        interval_start: Optional[timedelta] = None,
        interval_end: Optional[timedelta] = None,
        fill_selection: Optional[bool] = False,
    ):
        """
        Selects time slice.

        Either `history_duration` and `history_duration` or `interval_start` and
        `interval_end` must be supplied.

        Args:
            source_datapipe: Datapipe of Xarray objects
            t0_datapipe: Datapipe of t0 times
            sample_period_duration: Sample period of xarray data
            history_duration (optional): History time used
            forecast_duration (optional): Forecast time used
            interval_start (optional): timedelta with respect to t0 where the open interval begins
            interval_end (optional): timedelta with respect to t0 where the open interval ends
            fill_selection (optional): If True, and if the data yielded from `source_datapipe` does
                not extend over the entire requested time period. The missing timestamps are filled
                with NaN values in the returned xarray object. Else the default xarray slicing
                behaviour is used.
        """
        self.source_datapipe = source_datapipe
        self.t0_datapipe = t0_datapipe
        self.fill_selection = fill_selection

        used_duration = history_duration is not None and forecast_duration is not None
        used_intervals = interval_start is not None and interval_end is not None
        assert used_duration ^ used_intervals, "Either durations, or intervals must be supplied"

        if used_duration:
            self.interval_start = -np.timedelta64(history_duration)
            self.interval_end = np.timedelta64(forecast_duration)
        elif used_intervals:
            self.interval_start = np.timedelta64(interval_start)
            self.interval_end = np.timedelta64(interval_end)

        self.sample_period_duration = sample_period_duration

    def _sel_fillnan(self, xr_data, start_dt, end_dt):
        requested_times = pd.date_range(
            start_dt,
            end_dt,
            freq=self.sample_period_duration,
        )
        # Missing time indexes are returned with all NaN values
        return xr_data.reindex(time_utc=requested_times)

    def _sel_default(self, xr_data, start_dt, end_dt):
        return xr_data.sel(time_utc=slice(start_dt, end_dt))

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        xr_data = next(iter(self.source_datapipe))

        for t0 in self.t0_datapipe:
            t0_datetime_utc = pd.Timestamp(t0)
            start_dt = t0_datetime_utc + self.interval_start
            end_dt = t0_datetime_utc + self.interval_end

            start_dt = start_dt.ceil(self.sample_period_duration)
            end_dt = end_dt.ceil(self.sample_period_duration)

            if self.fill_selection:
                yield self._sel_fillnan(xr_data, start_dt, end_dt)
            else:
                yield self._sel_default(xr_data, start_dt, end_dt)



"""Select the t0 time for training"""
import logging
from typing import Optional

import numpy as np
import pandas as pd
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

logger = logging.getLogger(__name__)


@functional_datapipe("select_t0_time")
class SelectT0TimeIterDataPipe(IterDataPipe):
    """Select the random t0 time for the training data"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        dim_name: str = "time_utc",
        return_all_times: Optional[bool] = False,
        number_locations_datapipe: Optional[IterDataPipe] = None,
    ):
        """
        Select a random t0 time for training

        Args:
            source_datapipe: Datapipe emitting Xarray objects
            dim_name: The time dimension name to use
            return_all_times: option to
            number_locations_datapipe: get the total number of locations
        """
        self.source_datapipe = source_datapipe
        self.dim_name = dim_name
        self.return_all_times = return_all_times
        self.number_locations_datapipe = number_locations_datapipe

        if self.return_all_times:
            logger.debug("Will be returning all t0 times")

    def __iter__(self) -> pd.Timestamp:
        """Get the latest timestamp and return it"""
        for xr_data in self.source_datapipe:
            if self.return_all_times and (self.number_locations_datapipe is not None):
                number_of_locations = next(self.number_locations_datapipe)

                logger.info(
                    f"Will be returning all times from {xr_data[self.dim_name]}. "
                    f"There are {len(xr_data[self.dim_name])} of them"
                )

                for t0 in xr_data[self.dim_name].values:
                    for _ in range(0, number_of_locations):
                        logger.debug(f"t0 will be {t0}")
                        yield t0

            else:
                logger.debug(f"Selecting t0 from {len(xr_data[self.dim_name])} datetimes")

                if len(xr_data[self.dim_name].values) == 0:
                    raise Exception("There are no values to get t0 from")
                t0 = np.random.choice(xr_data[self.dim_name].values)
                logger.debug(f"t0 will be {t0}")

                yield t0



""" Data module for pytorch lightning """
from typing import Callable, Union

from pytorch_lightning import LightningDataModule
from torch.utils.data import DataLoader

from ocf_datapipes.batch.fake.fake_batch import fake_data_pipeline


class DataModule(LightningDataModule):
    """
    Example of LightningDataModule using ocf_datapipes
    """

    def __init__(
        self,
        data_pipeline: Union[str, Callable],
        configuration,
        fake_data,
        n_train_data,
        n_val_data,
    ):
        """
        Set up datamodule

        Args:
            data_pipeline: the pipeline function to be called,
                can also be the name of the data pipeline
            configuration: the configuration file
            fake_data: option to load fake data or not
            n_train_data: TODO
            n_val_data: TODO
        """

        super().__init__()
        self.data_pipeline = data_pipeline
        self.configuration = configuration
        self.fake_data = fake_data
        self.n_train_data = n_train_data
        self.n_val_data = n_val_data

        if self.fake_data:
            self.data_pipeline = fake_data_pipeline
        elif isinstance(self.data_pipeline, str):
            if self.data_pipeline == "pv_satellite_nwp":
                from ocf_datapipes.training.pv_satellite_nwp import pv_nwp_satellite_data_pipeline

                self.data_pipeline = pv_nwp_satellite_data_pipeline

            if self.data_pipeline == "gsp_pv_satellite_nwp":
                from ocf_datapipes.training.gsp_pv_satellite_nwp import (
                    gsp_pv_nwp_satellite_data_pipeline,
                )

                self.data_pipeline = gsp_pv_nwp_satellite_data_pipeline

    def train_dataloader(self):
        """Get the train dataloader"""

        train_dataset = self.data_pipeline(configuration=self.configuration)
        train_dataloader = DataLoader(train_dataset, batch_size=None)

        return train_dataloader

    def val_dataloader(self):
        """Get the validation dataloader"""

        validation_data_pipeline = self.data_pipeline(configuration=self.configuration)
        validation_dataloader = DataLoader(validation_data_pipeline, batch_size=None)

        return validation_dataloader

    def test_dataloader(self):
        """Get the test dataloader"""

        test_data_pipeline = self.data_pipeline(configuration=self.configuration)
        test_dataloader = DataLoader(test_data_pipeline, batch_size=None)

        return test_dataloader



"""Convert PV to Numpy Batch"""
import logging

import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float

logger = logging.getLogger(__name__)


@functional_datapipe("convert_pv_to_numpy_batch")
class ConvertPVToNumpyBatchIterDataPipe(IterDataPipe):
    """Convert PV Xarray to NumpyBatch"""

    def __init__(self, source_datapipe: IterDataPipe):
        """
        Convert PV Xarray objects to NumpyBatch objects

        Args:
            source_datapipe: Datapipe emitting PV Xarray objects
        """
        super().__init__()
        self.source_datapipe = source_datapipe

    def __iter__(self) -> NumpyBatch:
        """Iterate and convert PV Xarray to NumpyBatch"""
        for xr_data in self.source_datapipe:
            logger.debug("Converting PV xarray to numpy example")

            example: NumpyBatch = {
                BatchKey.pv: xr_data.values,
                BatchKey.pv_t0_idx: xr_data.attrs["t0_idx"],
                BatchKey.pv_system_row_number: xr_data["pv_system_row_number"].values,
                BatchKey.pv_id: xr_data["pv_system_id"].values.astype(np.float32),
                BatchKey.pv_capacity_watt_power: xr_data["capacity_watt_power"].values,
                BatchKey.pv_time_utc: datetime64_to_float(xr_data["time_utc"].values),
                BatchKey.pv_x_osgb: xr_data["x_osgb"].values,
                BatchKey.pv_y_osgb: xr_data["y_osgb"].values,
            }

            yield example



import numpy as np

from ocf_datapipes.transform.xarray import CreateTimeImage


def test_create_time_image_sat(sat_datapipe):
    time_image_datapipe = CreateTimeImage(sat_datapipe)
    data = next(iter(time_image_datapipe))
    assert 1.0 >= np.max(data)
    assert -1.0 <= np.min(data)


def test_create_time_image_hrv(sat_hrv_datapipe):
    time_image_datapipe = CreateTimeImage(sat_hrv_datapipe)
    data = next(iter(time_image_datapipe))
    assert 1.0 >= np.max(data)
    assert -1.0 <= np.min(data)



"""Convert Satellite to NumpyBatch"""
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch
from ocf_datapipes.utils.utils import datetime64_to_float


@functional_datapipe("convert_satellite_to_numpy_batch")
class ConvertSatelliteToNumpyBatchIterDataPipe(IterDataPipe):
    """Converts Xarray Satellite to NumpyBatch object"""

    def __init__(self, source_datapipe: IterDataPipe, is_hrv: bool = False):
        """
        Converts Xarray satellite object to NumpyBatch object

        Args:
            source_datapipe: Datapipe emitting Xarray satellite objects
            is_hrv: Whether this is HRV satellite data or non-HRV data
        """
        super().__init__()
        self.source_datapipe = source_datapipe
        self.is_hrv = is_hrv

    def __iter__(self) -> NumpyBatch:
        """Convert each example to a NumpyBatch object"""
        for xr_data in self.source_datapipe:
            if self.is_hrv:
                example: NumpyBatch = {
                    BatchKey.hrvsatellite_actual: xr_data.values,
                    BatchKey.hrvsatellite_t0_idx: xr_data.attrs["t0_idx"],
                    BatchKey.hrvsatellite_time_utc: datetime64_to_float(xr_data["time_utc"].values),
                }

                for batch_key, dataset_key in (
                    (BatchKey.hrvsatellite_y_geostationary, "y_geostationary"),
                    (BatchKey.hrvsatellite_x_geostationary, "x_geostationary"),
                ):
                    # HRVSatellite coords are already float32.
                    example[batch_key] = xr_data[dataset_key].values
            else:
                example: NumpyBatch = {
                    BatchKey.satellite_actual: xr_data.values,
                    BatchKey.satellite_t0_idx: xr_data.attrs["t0_idx"],
                    BatchKey.satellite_time_utc: datetime64_to_float(xr_data["time_utc"].values),
                }

                for batch_key, dataset_key in (
                    (BatchKey.satellite_y_geostationary, "y_geostationary"),
                    (BatchKey.satellite_x_geostationary, "x_geostationary"),
                ):
                    # HRVSatellite coords are already float32.
                    example[batch_key] = xr_data[dataset_key].values

            yield example



"""Create the training/validation datapipe for training the PVNet Model"""
import logging
from datetime import datetime, timedelta
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import xarray as xr
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.batch import MergeNumpyModalities
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenGSPFromDatabase
from ocf_datapipes.training.common import (
    create_t0_and_loc_datapipes,
    open_and_return_datapipes,
)
from ocf_datapipes.utils.consts import (
    NEW_NWP_MEAN,
    NEW_NWP_STD,
    RSS_MEAN,
    RSS_STD,
    BatchKey,
    NumpyBatch,
)

xr.set_options(keep_attrs=True)
logger = logging.getLogger("pvnet_datapipe")


def normalize_gsp(x):
    """Normalize the GSP data

    Args:
        x: Input DataArray

    Returns:
        Normalized DataArray
    """
    return x / x.capacity_megawatt_power


def production_sat_scale(x):
    """Scale the production satellite data

    Args:
        x: Input DataArray

    Returns:
        Scaled DataArray
    """
    return x / 1024


def pvnet_concat_gsp(gsp_dataarrays: List[xr.DataArray]):
    """This function is used to combine the split history and future gsp dataarrays.

    These are split inside the `slice_datapipes_by_time()` function below.

    Splitting them inside that function allows us to apply dropout to the
    history GSP whilst leaving the future GSP without NaNs.

    We recombine the history and future with this function to allow us to use the
    `MergeNumpyModalities()` datapipe without redefining the BatchKeys.

    The `pvnet` model was also written to use a GSP array which has historical and future
    and to split it out. These maintains that assumption.
    """
    return xr.concat(gsp_dataarrays, dim="time_utc")


def gsp_drop_national(x: Union[xr.DataArray, xr.Dataset]):
    """Drop entries for national PV output

    Args:
        x: Data source of gsp data

    Returns:
        Filtered data source
    """
    return x.where(x.gsp_id != 0, drop=True)


def fill_nans_in_arrays(batch: NumpyBatch) -> NumpyBatch:
    """Fills all NaN values in each np.ndarray in the batch dictionary with zeros.

    Operation is performed in-place on the batch.
    """
    logger.info("Filling Nans with zeros")
    for k, v in batch.items():
        if isinstance(v, np.ndarray):
            np.nan_to_num(v, copy=False, nan=0.0)
    return batch


class AddZeroedSatelliteData:
    """A callable class used to add zeroed-out satellite data to batches of data.

    This is useful
    to speed up batch loading if pre-training the output part of the network without satellite
    inputs.
    """

    def __init__(self, configuration: Configuration, is_hrv: bool = False):
        """A callable class used to add zeroed-out satellite data to batches of data.

        Args:
            configuration: Configuration object
            is_hrv: If False, non-HRV data is added by called function, else HRV.
        """

        self.configuration = configuration
        self.is_hrv = is_hrv

    def __call__(self, batch: NumpyBatch) -> NumpyBatch:
        """Add zeroed-out satellite data to batch with shape accoriding to supplied configuration.

        Batch is modified in-place and returned.

        Args:
            batch: Numpy batch of input data.
        """

        variable = "hrvsatellite" if self.is_hrv else "satellite"

        satellite_config = getattr(self.configuration.input_data, variable)

        n_channels = len(getattr(satellite_config, f"{variable}_channels"))
        height = getattr(satellite_config, f"{variable}_image_size_pixels_height")
        width = getattr(satellite_config, f"{variable}_image_size_pixels_width")

        sequence_len = satellite_config.history_minutes // 5 + 1 - 3

        batch[getattr(BatchKey, f"{variable}_actual")] = np.zeros(
            (sequence_len, n_channels, height, width)
        )

        return batch


class AddZeroedNWPData:
    """A callable class used to add zeroed-out NWP data to batches of data.

    This is useful to speed up batch loading if pre-training the output part of the network without
    NWP inputs.
    """

    def __init__(self, configuration: Configuration):
        """A callable class used to add zeroed-out NWP data to batches of data.

        Args:
            configuration: Configuration object
        """
        self.configuration = configuration

    def __call__(self, batch: NumpyBatch) -> NumpyBatch:
        """Add zeroed-out NWP data to batch with shape accoriding to supplied configuration.

        Batch is modified in-place and returned.

        Args:
            batch: Numpy batch of input data.
        """

        config = self.configuration.input_data.nwp

        n_channels = len(config.nwp_channels)
        height = config.nwp_image_size_pixels_height
        width = config.nwp_image_size_pixels_width

        sequence_len = config.history_minutes // 60 + config.forecast_minutes // 60 + 1

        batch[BatchKey.nwp] = np.zeros((sequence_len, n_channels, height, width))

        return batch


class DatapipeKeyForker:
    """ "Internal helper function to track forking of a datapipe."""

    def __init__(self, keys: List, datapipe: IterDataPipe):
        """Internal helper function to track forking of a datapipe.

        As forks are returned, this object tracks the keys left and returns the final copy of the
        datapipe when the last key is requested. This makes multiple forking easier and ensures
        closure.

        Args:
            keys: List of keys for which datapipe duplication is required.
            datapipe: Datapipe which will be forked for each ket
        """
        self.keys_left = keys
        self.datapipe = datapipe

    def __call__(self, key):
        """ "Returns a fork of `self.datapipe` and tracks a the keys left to ensure closure.

        Args:
            key: key to remove from `self.keys_left`. If `key` is None then an extra copy is made
            without affecting `self.keys_left`.
        """
        if len(self.keys_left) == 0:
            raise ValueError(f"No keys left when requested key : {key}")
        if key is not None:
            self.keys_left.remove(key)
        if len(self.keys_left) > 0:
            self.datapipe, return_datapipe = self.datapipe.fork(2, buffer_size=5)
        else:
            return_datapipe = self.datapipe
        return return_datapipe

    def close(self):
        """Asserts that the keys have all been used."""
        assert len(self.keys_left) == 0


def _get_datapipes_dict(
    config_filename: str,
    block_sat: bool,
    block_nwp: bool,
    production: bool = False,
):
    # Load datasets
    datapipes_dict = open_and_return_datapipes(
        configuration_filename=config_filename,
        use_gsp=(not production),
        use_pv=False,
        use_sat=not block_sat,  # Only loaded if we aren't replacing them with zeros
        use_hrv=False,
        use_nwp=not block_nwp,  # Only loaded if we aren't replacing them with zeros
        use_topo=False,
    )
    if production:
        configuration: Configuration = datapipes_dict["config"]

        datapipes_dict["gsp"] = OpenGSPFromDatabase().add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=30),
            history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        )
        datapipes_dict["sat"] = datapipes_dict["sat"].map(production_sat_scale)

    return datapipes_dict


def construct_loctime_pipelines(
    config_filename: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    block_sat: bool = False,
    block_nwp: bool = False,
) -> Tuple[IterDataPipe, IterDataPipe]:
    """Construct location and time pipelines for the input data config file.

    Args:
        config_filename: Path to config file.
        start_time: Minimum time for time datapipe.
        end_time: Maximum time for time datapipe.
        block_sat: Whether to load zeroes for satellite data.
        block_nwp: Whether to load zeroes for NWP data.
    """

    datapipes_dict = _get_datapipes_dict(
        config_filename,
        block_sat,
        block_nwp,
    )

    # Pull out config file
    config = datapipes_dict.pop("config")

    # We sample time and space of other data using GSP time and space coordinates, so filter GSP
    # data first amd this is carried through
    datapipes_dict["gsp"] = datapipes_dict["gsp"].map(gsp_drop_national)
    if (start_time is not None) or (end_time is not None):
        datapipes_dict["gsp"] = datapipes_dict["gsp"].select_train_test_time(start_time, end_time)

    # Get overlapping time periods
    location_pipe, t0_datapipe = create_t0_and_loc_datapipes(
        datapipes_dict,
        configuration=config,
        key_for_t0="gsp",
        shuffle=True,
        nwp_max_t0_offset=minutes(180),
    )

    return location_pipe, t0_datapipe


def minutes(num_mins: int):
    """Timedelta of a number of minutes.

    Args:
        num_mins: Minutes timedelta.
    """
    return timedelta(minutes=num_mins)


def slice_datapipes_by_time(
    datapipes_dict: Dict,
    t0_datapipe: IterDataPipe,
    configuration: Configuration,
    production: bool = False,
) -> None:
    """
    Modifies a dictionary of datapipes in-place to yield samples for given times t0.

    The NWP data* will be at least 90 minutes stale (i.e. as if it takes 90 minutes for the foreast
    to become available).

    The satellite data* is shaped so that the most recent can be 15 minutes before t0. However, 50%
    of the time dropout is applied so that the most recent field is between 45 and 20 minutes before
    t0. When dropped out like this, the values after this selected dropout time are set to NaN.

    The HRV data* is similar to the satellite data and if both are included they drop out
    simulataneously.

    The GSP data is split into "gsp" and "gsp_future" keys. 10% of the time the gsp value for time
    t0, which occurs under the "gsp" key, is set to NaN

    The PV data* is also split it "pv" and "pv_future" keys.

    * if included

    n.b. PV and HRV are included in this function, but not yet in the rest of the pvnet pipeline.
    This is mostly for demonstratio purposes of how the dropout might be applied.

    Args:
        datapipes_dict: Dictionary of used datapipes and t0 ones
        t0_datapipe: Datapipe which yields t0 times for sample
        configuration: Configuration object.
        production: Whether constucting pipeline for production inference. No dropout is used if
            True.

    """

    conf_in = configuration.input_data

    # Use DatapipeKeyForker to avoid forking t0_datapipe too many times, or leaving any forks unused
    fork_keys = {k for k in datapipes_dict.keys() if k not in ["topo"]}
    get_t0_datapipe = DatapipeKeyForker(fork_keys, t0_datapipe)

    sat_and_hrv_dropout_kwargs = dict(
        # Satellite is either 30 minutes or 60 minutes delayed
        dropout_timedeltas=[minutes(-60), minutes(-30)],
        dropout_frac=0 if production else 1.0,
    )

    # Satellite data never more recent than t0-30mins
    if production:
        sat_delay = minutes(-configuration.input_data.satellite.live_delay_minutes)
    else:
        sat_delay = minutes(-30)

    if "nwp" in datapipes_dict:
        datapipes_dict["nwp"] = datapipes_dict["nwp"].convert_to_nwp_target_time_with_dropout(
            t0_datapipe=get_t0_datapipe("nwp"),
            sample_period_duration=minutes(60),
            history_duration=minutes(conf_in.nwp.history_minutes),
            forecast_duration=minutes(conf_in.nwp.forecast_minutes),
            # The NWP forecast will always be at least 180 minutes stale
            dropout_timedeltas=[minutes(-180)],
            dropout_frac=0 if production else 1.0,
        )

    if "sat" in datapipes_dict:
        # Take time slices of sat data
        datapipes_dict["sat"] = datapipes_dict["sat"].select_time_slice(
            t0_datapipe=get_t0_datapipe(None),
            sample_period_duration=minutes(5),
            interval_start=minutes(-conf_in.satellite.history_minutes),
            interval_end=sat_delay,
            fill_selection=production,
        )

        # Generate randomly sampled dropout times
        sat_dropout_time_datapipe = get_t0_datapipe("sat").select_dropout_time(
            **sat_and_hrv_dropout_kwargs
        )

        if "hrv" in datapipes_dict:
            # Make dropout-time copy for hrv if included in data.
            # HRV and non-HRV will dropout simultaneously.
            sat_dropout_time_datapipe, hrv_dropout_time_datapipe = sat_dropout_time_datapipe.fork(
                2, buffer_size=5
            )

        # Apply the dropout
        datapipes_dict["sat"] = datapipes_dict["sat"].apply_dropout_time(
            dropout_time_datapipe=sat_dropout_time_datapipe,
        )

    if "hrv" in datapipes_dict:
        if "sat" not in datapipes_dict:
            # Generate randomly sampled dropout times
            # This is shared with sat if sat included
            hrv_dropout_time_datapipe = get_t0_datapipe(None).select_dropout_time(
                **sat_and_hrv_dropout_kwargs
            )

        datapipes_dict["hrv"] = datapipes_dict["hrv"].select_time_slice(
            t0_datapipe=get_t0_datapipe("hrv"),
            sample_period_duration=minutes(5),
            interval_start=minutes(-conf_in.hrvsatellite.history_minutes),
            interval_end=sat_delay,
            fill_selection=production,
        )

        # Apply the dropout
        datapipes_dict["hrv"] = datapipes_dict["hrv"].apply_dropout_time(
            dropout_time_datapipe=hrv_dropout_time_datapipe,
        )

    if "pv" in datapipes_dict:
        datapipes_dict["pv"], dp = datapipes_dict["pv"].fork(2, buffer_size=5)

        datapipes_dict["pv_future"] = dp.select_time_slice(
            t0_datapipe=get_t0_datapipe(None),
            sample_period_duration=minutes(5),
            interval_start=minutes(5),
            interval_end=minutes(conf_in.pv.forecast_minutes),
            fill_selection=production,
        )

        datapipes_dict["pv"] = datapipes_dict["pv"].select_time_slice(
            t0_datapipe=get_t0_datapipe("pv"),
            sample_period_duration=minutes(5),
            interval_start=minutes(-conf_in.pv.history_minutes),
            interval_end=minutes(0),
            fill_selection=production,
        )

    if "gsp" in datapipes_dict:
        datapipes_dict["gsp"], dp = datapipes_dict["gsp"].fork(2, buffer_size=5)

        datapipes_dict["gsp_future"] = dp.select_time_slice(
            t0_datapipe=get_t0_datapipe(None),
            sample_period_duration=minutes(30),
            interval_start=minutes(30),
            interval_end=minutes(conf_in.gsp.forecast_minutes),
            fill_selection=production,
        )

        datapipes_dict["gsp"] = datapipes_dict["gsp"].select_time_slice(
            t0_datapipe=get_t0_datapipe(None),
            sample_period_duration=minutes(30),
            interval_start=-minutes(conf_in.gsp.history_minutes),
            interval_end=minutes(0),
            fill_selection=production,
        )

        # Dropout on the GSP, but not the future GSP
        gsp_dropout_time_datapipe = get_t0_datapipe("gsp").select_dropout_time(
            # GSP data for time t0 may be missing. Only have value for t0-30mins
            dropout_timedeltas=[minutes(-30)],
            dropout_frac=0 if production else 0.1,
        )

        datapipes_dict["gsp"] = datapipes_dict["gsp"].apply_dropout_time(
            dropout_time_datapipe=gsp_dropout_time_datapipe,
        )

    get_t0_datapipe.close()

    return


def construct_sliced_data_pipeline(
    config_filename: str,
    location_pipe: IterDataPipe,
    t0_datapipe: IterDataPipe,
    block_sat: bool = False,
    block_nwp: bool = False,
    production: bool = False,
    check_satellite_no_zeros: bool = False,
) -> IterDataPipe:
    """Constructs data pipeline for the input data config file.

    This yields samples from the location and time datapipes.

    Args:
        config_filename: Path to config file.
        location_pipe: Datapipe yielding locations.
        t0_datapipe: Datapipe yielding times.
        block_sat: Whether to load zeroes for satellite data.
        block_nwp: Whether to load zeroes for NWP data.
        production: Whether constucting pipeline for production inference.
        check_satellite_no_zeros: Whether to check that satellite data has no zeros.
    """

    assert not (production and (block_sat or block_nwp))

    datapipes_dict = _get_datapipes_dict(
        config_filename,
        block_sat,
        block_nwp,
        production=production,
    )

    configuration = datapipes_dict.pop("config")

    # Unpack for convenience
    conf_sat = configuration.input_data.satellite
    conf_nwp = configuration.input_data.nwp

    # Slice all of the datasets by time - this is an in-place operation
    slice_datapipes_by_time(datapipes_dict, t0_datapipe, configuration, production)

    # Spatially slice, normalize, and convert data to numpy arrays
    numpy_modalities = []

    if "nwp" in datapipes_dict:
        nwp_datapipe = datapipes_dict["nwp"]

        location_pipe, location_pipe_copy = location_pipe.fork(2, buffer_size=5)
        nwp_datapipe = nwp_datapipe.select_spatial_slice_pixels(
            location_pipe_copy,
            roi_height_pixels=conf_nwp.nwp_image_size_pixels_height,
            roi_width_pixels=conf_nwp.nwp_image_size_pixels_width,
            x_dim_name="x_osgb",
            y_dim_name="y_osgb",
        )
        nwp_datapipe = nwp_datapipe.normalize(mean=NEW_NWP_MEAN, std=NEW_NWP_STD)
        numpy_modalities.append(nwp_datapipe.convert_nwp_to_numpy_batch())

    if "sat" in datapipes_dict:
        sat_datapipe = datapipes_dict["sat"]

        location_pipe, location_pipe_copy = location_pipe.fork(2, buffer_size=5)
        sat_datapipe = sat_datapipe.select_spatial_slice_pixels(
            location_pipe_copy,
            roi_height_pixels=conf_sat.satellite_image_size_pixels_height,
            roi_width_pixels=conf_sat.satellite_image_size_pixels_width,
            x_dim_name="x_geostationary",
            y_dim_name="y_geostationary",
        )
        sat_datapipe = sat_datapipe.normalize(mean=RSS_MEAN, std=RSS_STD)
        numpy_modalities.append(sat_datapipe.convert_satellite_to_numpy_batch())

    # GSP always assumed to be in data
    location_pipe, location_pipe_copy = location_pipe.fork(2, buffer_size=5)
    gsp_future_datapipe = datapipes_dict["gsp_future"]
    gsp_future_datapipe = gsp_future_datapipe.select_spatial_slice_meters(
        location_datapipe=location_pipe_copy,
        roi_height_meters=1,
        roi_width_meters=1,
        y_dim_name="y_osgb",
        x_dim_name="x_osgb",
        dim_name="gsp_id",
    )

    gsp_datapipe = datapipes_dict["gsp"]
    gsp_datapipe = gsp_datapipe.select_spatial_slice_meters(
        location_datapipe=location_pipe,
        roi_height_meters=1,
        roi_width_meters=1,
        y_dim_name="y_osgb",
        x_dim_name="x_osgb",
        dim_name="gsp_id",
    )

    # Recombine GSP arrays - see function doc for further explanation
    gsp_datapipe = gsp_datapipe.zip_ocf(gsp_future_datapipe).map(pvnet_concat_gsp)
    gsp_datapipe = gsp_datapipe.normalize(normalize_fn=normalize_gsp)

    numpy_modalities.append(gsp_datapipe.convert_gsp_to_numpy_batch())

    logger.debug("Combine all the data sources")
    combined_datapipe = MergeNumpyModalities(numpy_modalities).add_sun_position(modality_name="gsp")

    if block_sat and conf_sat != "":
        sat_block_func = AddZeroedSatelliteData(configuration)
        combined_datapipe = combined_datapipe.map(sat_block_func)

    if block_nwp and conf_nwp != "":
        nwp_block_func = AddZeroedNWPData(configuration)
        combined_datapipe = combined_datapipe.map(nwp_block_func)

    logger.info("Filtering out samples with no data")
    if check_satellite_no_zeros:
        # in production we don't want any nans in the satellite data
        combined_datapipe = combined_datapipe.map(check_nans_in_satellite_data)

    combined_datapipe = combined_datapipe.map(fill_nans_in_arrays)

    return combined_datapipe


def pvnet_datapipe(
    config_filename: str,
    start_time: Optional[datetime] = None,
    end_time: Optional[datetime] = None,
    block_sat: bool = False,
    block_nwp: bool = False,
) -> IterDataPipe:
    """
    Construct pvnet pipeline for the input data config file.

    Args:
        config_filename: Path to config file.
        start_time: Minimum time at which a sample can be selected.
        end_time: Maximum time at which a sample can be selected.
        block_sat: Whether to load zeroes for satellite data.
        block_nwp: Whether to load zeroes for NWP data.
    """
    logger.info("Constructing pvnet pipeline")

    # Open datasets from the config and filter to useable location-time pairs
    location_pipe, t0_datapipe = construct_loctime_pipelines(
        config_filename,
        start_time,
        end_time,
        block_sat,
        block_nwp,
    )

    # Shard after we have the loc-times. These are already shuffled so no need to shuffle again
    location_pipe = location_pipe.sharding_filter()
    t0_datapipe = t0_datapipe.sharding_filter()

    # In this function we re-open the datasets to make a clean separation before/after sharding
    # This function
    datapipe = construct_sliced_data_pipeline(
        config_filename,
        location_pipe,
        t0_datapipe,
        block_sat,
        block_nwp,
    )

    return datapipe


def check_nans_in_satellite_data(batch: NumpyBatch) -> NumpyBatch:
    """
    Check if there are any Nans values in the satellite data.
    """
    if np.any(np.isnan(batch[BatchKey.satellite_actual])):
        logger.error("Found nans values in satellite data")
        raise ValueError("Found nans values in satellite data")

    return batch



"""Create image channels of time values at each pixel"""
from typing import Union

import numpy as np
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.utils import trigonometric_datetime_transformation


@functional_datapipe("create_time_image")
class CreateTimeImageIterDataPipe(IterDataPipe):
    """Create Time image"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        image_dim: str = "geostationary",
        time_dim: str = "time_utc",
    ):
        """
        Creates a 3D data cube of Time output image x number of timesteps

        Args:
            source_datapipe: Source datapipe of PV data
            image_dim: Dimension name for the x and y dimensions
            time_dim: Time dimension name
        """
        self.source_datapipe = source_datapipe
        self.x_dim = "x_" + image_dim
        self.y_dim = "y_" + image_dim
        self.time_dim = time_dim

    def __iter__(self) -> xr.DataArray:
        for image_xr in self.source_datapipe:
            # Create empty image to use, assumes image has x and y coordinates
            time_image = _create_time_image(
                image_xr,
                time_dim=self.time_dim,
                output_width_pixels=len(image_xr[self.x_dim]),
                output_height_pixels=len(image_xr[self.y_dim]),
            )
            time_image = _create_data_array_from_image(
                time_image,
                image_xr,
                is_geostationary="x_geostationary" in image_xr.dims,
                time_dim=self.time_dim,
            )
            yield time_image


def _create_time_image(xr_data, time_dim: str, output_height_pixels: int, output_width_pixels: int):
    # Create trig decomposition of datetime values, tiled over output height and width
    datetimes = xr_data[time_dim].values
    trig_decomposition = trigonometric_datetime_transformation(datetimes)
    tiled_data = np.expand_dims(trig_decomposition, (2, 3))
    tiled_data = np.tile(tiled_data, (1, 1, output_height_pixels, output_width_pixels))
    return tiled_data


def _create_data_array_from_image(
    time_image: np.ndarray,
    image_xr: Union[xr.Dataset, xr.DataArray],
    is_geostationary: bool,
    time_dim: str,
):
    if is_geostationary:
        data_array = xr.DataArray(
            data=time_image,
            coords=(
                ("time_utc", image_xr[time_dim].values),
                (
                    "channel",
                    [f"time_channel_{i}" for i in range(time_image.shape[1])],
                ),  # Temp channel names
                ("y_geostationary", image_xr.y_geostationary.values),
                ("x_geostationary", image_xr.x_geostationary.values),
            ),
            name="time_image",
        ).astype(np.float32)
    else:
        data_array = xr.DataArray(
            data=time_image,
            coords=(
                ("time_utc", image_xr[time_dim].values),
                ("channel", [f"time_channel_{i}" for i in range(time_image.shape[1])]),
                ("y_osgb", image_xr.y_osgb.values),
                ("x_osgb", image_xr.x_osgb.values),
            ),
            name="time_image",
        ).astype(np.float32)
    data_array.attrs = image_xr.attrs
    return data_array



"""Create image channels of sun position at each pixel"""
from typing import Union

import numpy as np
import pvlib
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import Location
from ocf_datapipes.utils.geospatial import (
    load_geostationary_area_definition_and_transform_latlon,
    load_geostationary_area_definition_and_transform_osgb,
    osgb_to_lat_lon,
)

ELEVATION_MEAN = 37.4
ELEVATION_STD = 12.7
AZIMUTH_MEAN = 177.7
AZIMUTH_STD = 41.7


@functional_datapipe("create_sun_image")
class CreateSunImageIterDataPipe(IterDataPipe):
    """Create Sun image from individual sites"""

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        normalize: bool = False,
        image_dim: str = "geostationary",
        time_dim: str = "time_utc",
    ):
        """
        Creates a 3D data cube of PV output image x number of timesteps

        TODO Include PV System IDs or something, currently just takes the outputs

        Args:
            source_datapipe: Source datapipe of PV data
            normalize: Whether to normalize based off the image max, or leave raw data
            image_dim: Dimension name for the x and y dimensions
            time_dim: Time dimension name
        """
        self.source_datapipe = source_datapipe
        self.normalize = normalize
        self.x_dim = "x_" + image_dim
        self.y_dim = "y_" + image_dim
        self.time_dim = time_dim

    def __iter__(self) -> xr.DataArray:
        for image_xr in self.source_datapipe:
            # Create empty image to use for the PV Systems, assumes image has x and y coordinates
            sun_image = np.zeros(
                (
                    len(image_xr[self.time_dim]),
                    2,  # Azimuth and elevation
                    len(image_xr[self.y_dim]),
                    len(image_xr[self.x_dim]),
                ),
                dtype=np.float32,
            )
            if "geostationary" in self.x_dim:
                transform_to_latlon = load_geostationary_area_definition_and_transform_latlon(
                    image_xr
                )
                lats, lons = transform_to_latlon(
                    xx=image_xr[self.x_dim].values, yy=image_xr[self.y_dim].values
                )
                print(lats.shape)

            else:
                transform_to_latlon = osgb_to_lat_lon
                lats, lons = transform_to_latlon(x=image_xr.x_osgb.values, y=image_xr.y_osgb.values)
            time_utc = image_xr[self.time_dim].values

            # Loop round each example to get the Sun's elevation and azimuth:
            # Go through each time on its own, lat lons still in order of image
            # TODO Make this faster
            # dt = pd.DatetimeIndex(dt)  # pvlib expects a `pd.DatetimeIndex`.
            for y_index, lat in enumerate(lats):
                for x_index, lon in enumerate(lons):
                    solpos = pvlib.solarposition.get_solarposition(
                        time=time_utc,
                        latitude=lat,
                        longitude=lon,
                        # Which `method` to use?
                        # pyephem seemed to be a good mix between speed
                        # and ease but causes segfaults!
                        # nrel_numba doesn't work when using
                        # multiple worker processes.
                        # nrel_c is probably fastest but
                        # requires C code to be manually compiled:
                        # https://midcdmz.nrel.gov/spa/
                    )
                    sun_image[:, 0, y_index, x_index] = solpos["azimuth"]
                    sun_image[:, 1, y_index, x_index] = solpos["elevation"]

            # Normalize.
            if self.normalize:
                sun_image[:, 0] = (sun_image[:, 0] - AZIMUTH_MEAN) / AZIMUTH_STD
                sun_image[:, 1] = (sun_image[:, 1] - ELEVATION_MEAN) / ELEVATION_STD

            # Should return Xarray as in Xarray transforms
            # Same coordinates as the image xarray, so can take that
            sun_image = _create_data_array_from_image(
                sun_image,
                image_xr,
                is_geostationary="geostationary" in self.x_dim,
                time_dim=self.time_dim,
            )
            yield sun_image


def _create_data_array_from_image(
    sun_image: np.ndarray,
    image_xr: Union[xr.Dataset, xr.DataArray],
    is_geostationary: bool,
    time_dim: str,
):
    if is_geostationary:
        data_array = xr.DataArray(
            data=sun_image,
            coords=(
                ("time_utc", image_xr[time_dim].values),
                ("channel", ["azimuth", "elevation"]),
                ("y_geostationary", image_xr.y_geostationary.values),
                ("x_geostationary", image_xr.x_geostationary.values),
            ),
            name="sun_image",
        ).astype(np.float32)
    else:
        data_array = xr.DataArray(
            data=sun_image,
            coords=(
                ("time_utc", image_xr[time_dim].values),
                ("channel", ["azimuth", "elevation"]),
                ("y_osgb", image_xr.y_osgb.values),
                ("x_osgb", image_xr.x_osgb.values),
            ),
            name="sun_image",
        ).astype(np.float32)
    data_array.attrs = image_xr.attrs
    return data_array


def _get_idx_of_pixel_closest_to_poi_geostationary(
    xr_data: xr.DataArray,
    center_osgb: Location,
    x_dim_name="x_geostationary",
    y_dim_name="y_geostationary",
) -> Location:
    """
    Return x and y index location of pixel at center of region of interest.

    Args:
        xr_data: Xarray dataset
        center_osgb: Center in OSGB coordinates
        x_dim_name: X dimension name
        y_dim_name: Y dimension name

    Returns:
        Location for the center pixel in geostationary coordinates
    """
    _osgb_to_geostationary = load_geostationary_area_definition_and_transform_osgb(xr_data)
    center_geostationary_tuple = _osgb_to_geostationary(xx=center_osgb.x, yy=center_osgb.y)
    center_geostationary = Location(
        x=center_geostationary_tuple[0], y=center_geostationary_tuple[1]
    )

    # Get the index into x and y nearest to x_center_geostationary and y_center_geostationary:
    x_index_at_center = np.searchsorted(xr_data[x_dim_name].values, center_geostationary.x) - 1
    # y_geostationary is in descending order:
    y_index_at_center = len(xr_data[y_dim_name]) - (
        np.searchsorted(xr_data[y_dim_name].values[::-1], center_geostationary.y) - 1
    )
    return Location(x=x_index_at_center, y=y_index_at_center)



"""Extends timestamps into the future"""
import logging

import numpy as np
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.consts import BatchKey, NumpyBatch

logger = logging.getLogger(__name__)


@functional_datapipe("extend_timesteps_to_future")
class ExtendTimestepsToFutureIterDataPipe(IterDataPipe):
    """Extends timestamps into the future"""

    def __init__(self, source_datapipe: IterDataPipe, forecast_duration, sample_period_duration):
        """
        Extends timestamps into the future

        This assumes that the current time_utc array only covers history + now,
        so just extends it further into the future

        Args:
            source_datapipe: Datapipe of NumpyBatch
            forecast_duration: Forecast duration time
            sample_period_duration: Sample period for forecast
        """
        self.source_datapipe = source_datapipe
        self.forecast_duration = forecast_duration
        self.sample_period_duration = sample_period_duration
        self.num_future_timesteps = int(self.forecast_duration / self.sample_period_duration)

    def __iter__(self) -> NumpyBatch:
        for np_batch in self.source_datapipe:
            all_time_dims: dict[BatchKey, np.ndarray] = {
                key: value for key, value in np_batch.items() if key.name.endswith("time_utc")
            }  # NWP should already cover the future, ones with no future are GSP, PV, and HRV
            # Take the current past ones and just append more to the future
            for time_dim_key in all_time_dims.keys():
                logger.debug(f"Extending {time_dim_key} for {self.num_future_timesteps} steps")
                # Get last time stamp = now
                times = list(all_time_dims[time_dim_key])
                timestep_diff = abs(times[-1] - times[-2])
                # Assumes they are all the same distance apart, just then add to itself n many times
                for future_step in range(self.num_future_timesteps):
                    times.append(times[-1] + timestep_diff)
                # Now re-numpify it and update the original
                times = np.asarray(times, dtype=np.float64)
                all_time_dims[time_dim_key] = times
            np_batch.update(all_time_dims)
            yield np_batch



"""Xarray transforms"""
from ocf_datapipes.transform.xarray.pv.ensure_n_pv_systems_per_example import (
    EnsureNPVSystemsPerExampleIterDataPipe as EnsureNPVSystemsPerExample,
)
from ocf_datapipes.transform.xarray.pv.pv_fill_nighttime_nans_with_zeros import (
    PVFillNightNansIterDataPipe as PVFillNightNans,
)
from ocf_datapipes.transform.xarray.pv.pv_power_rolling_window import (
    PVPowerRollingWindowIterDataPipe as PVPowerRollingWindow,
)
from ocf_datapipes.transform.xarray.pv.remove_pv_zero_examples import (
    PVPowerRemoveZeroDataIterDataPipe as PVPowerRemoveZeroData,
)

from .add_t0idx_and_sample_period_duration import (
    AddT0IdxAndSamplePeriodDurationIterDataPipe as AddT0IdxAndSamplePeriodDuration,
)
from .convert_satellite_to_int import ConvertSatelliteToInt8IterDataPipe as ConvertSatelliteToInt8
from .convert_to_numpy_stack import StackXarrayIterDataPipe as StackXarray
from .convert_to_nwp_target_times import (
    ConvertToNWPTargetTimeIterDataPipe as ConvertToNWPTargetTime,
)
from .create_sun_image import CreateSunImageIterDataPipe as CreateSunImage
from .create_time_image import CreateTimeImageIterDataPipe as CreateTimeImage
from .downsample import DownsampleIterDataPipe as Downsample
from .get_contiguous_time_periods import (
    GetContiguousT0TimePeriodsIterDataPipe as GetContiguousT0TimePeriods,
)
from .gsp.create_gsp_image import CreateGSPImageIterDataPipe as CreateGSPImage
from .gsp.ensure_n_gsp_per_example import (
    EnsureNGSPSPerExampleIterDataPipe as EnsureNGSPSPerExampleIter,
)
from .gsp.remove_northern_gsp import RemoveNorthernGSPIterDataPipe as RemoveNorthernGSP
from .metnet_preprocessor import PreProcessMetNetIterDataPipe as PreProcessMetNet
from .normalize import NormalizeIterDataPipe as Normalize
from .nwp_dropout import (
    ConvertToNWPTargetTimeWithDropoutIterDataPipe as ConvertToNWPTargetTimeWithDropout,
)
from .pv.assign_daynight_status import AssignDayNightStatusIterDataPipe as AssignDayNightStatus
from .pv.create_pv_history_image import CreatePVHistoryImageIterDataPipe as CreatePVHistoryImage
from .pv.create_pv_image import CreatePVImageIterDataPipe as CreatePVImage
from .pv.create_pv_meta_image import CreatePVMetadataImageIterDataPipe as CreatePVMetadataImage
from .remove_nans import RemoveNansIterDataPipe as RemoveNans
from .reproject_topographic_data import ReprojectTopographyIterDataPipe as ReprojectTopography
from .standard_dropout import ApplyDropoutTimeIterDataPipe as ApplyDropoutTime
from .standard_dropout import SelectDropoutTimeIterDataPipe as SelectDropoutTime



"""Common functionality for datapipes"""
import logging
from datetime import timedelta

from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import (
    OpenConfiguration,
    OpenGSP,
    OpenNWP,
    OpenPVFromNetCDF,
    OpenSatellite,
    OpenTopography,
)

logger = logging.getLogger(__name__)


def open_and_return_datapipes(
    configuration_filename: str,
    use_gsp: bool = True,
    use_nwp: bool = True,
    use_pv: bool = True,
    use_sat: bool = True,
    use_hrv: bool = True,
    use_topo: bool = True,
    production: bool = False,
) -> dict[str, IterDataPipe]:
    """
    Open data sources given a configuration and return the list of datapipes

    Args:
        configuration_filename: Path to file to open
        use_nwp: Whether to use NWP data or not
        use_topo: Whether to use topographic data
        use_pv: Whether to open PV data
        use_hrv: Whether to open HRV satellite data
        use_sat: Whether to open non-HRV satellite data
        use_gsp: Whether to use GSP data or not
        production: bool if this is for production or not

    Returns:
        List of datapipes corresponding to the datapipes to open
    """
    # load configuration
    config_datapipe = OpenConfiguration(configuration_filename)
    configuration: Configuration = next(iter(config_datapipe))

    # Check which modalities to use
    conf_in = configuration.input_data
    use_nwp = use_nwp and (conf_in.nwp.nwp_zarr_path != "")
    use_pv = use_pv and (conf_in.pv.pv_files_groups[0].pv_filename != "")
    use_sat = use_sat and (conf_in.satellite.satellite_zarr_path != "")
    use_hrv = use_hrv and (conf_in.hrvsatellite.hrvsatellite_zarr_path != "")
    use_topo = use_topo and (conf_in.topographic.topographic_filename != "")
    use_gsp = use_gsp and (conf_in.gsp.gsp_zarr_path != "")

    logger.debug(
        f"GSP: {use_gsp} NWP: {use_nwp} Sat: {use_sat},"
        f" HRV: {use_hrv} PV: {use_pv} Topo: {use_topo}"
    )

    used_datapipes = {}

    used_datapipes["config"] = configuration

    # Load GSP national data
    if use_gsp:
        logger.debug("Opening GSP Data")
        gsp_datapipe = OpenGSP(
            gsp_pv_power_zarr_path=configuration.input_data.gsp.gsp_zarr_path
        ).add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=30),
            history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
        )

        used_datapipes["gsp"] = gsp_datapipe

    # Load NWP data
    if use_nwp:
        logger.debug("Opening NWP Data")
        nwp_datapipe = (
            OpenNWP(configuration.input_data.nwp.nwp_zarr_path)
            .select_channels(configuration.input_data.nwp.nwp_channels)
            .add_t0_idx_and_sample_period_duration(
                sample_period_duration=timedelta(hours=1),
                history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
            )
        )

        used_datapipes["nwp"] = nwp_datapipe

    if use_sat:
        logger.debug("Opening Satellite Data")
        sat_datapipe = (
            OpenSatellite(
                configuration.input_data.satellite.satellite_zarr_path,
                use_15_minute_data_if_needed=production,
            )
            .select_channels(configuration.input_data.satellite.satellite_channels)
            .add_t0_idx_and_sample_period_duration(
                sample_period_duration=timedelta(minutes=5),
                history_duration=timedelta(
                    minutes=configuration.input_data.satellite.history_minutes
                ),
            )
        )

        used_datapipes["sat"] = sat_datapipe

    if use_hrv:
        logger.debug("Opening HRV Satellite Data")
        sat_hrv_datapipe = OpenSatellite(
            configuration.input_data.hrvsatellite.hrvsatellite_zarr_path
        ).add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(
                minutes=configuration.input_data.hrvsatellite.history_minutes
            ),
        )

        used_datapipes["hrv"] = sat_hrv_datapipe

    if use_pv:
        logger.debug("Opening PV")
        pv_datapipe = OpenPVFromNetCDF(
            pv=configuration.input_data.pv
        ).add_t0_idx_and_sample_period_duration(
            sample_period_duration=timedelta(minutes=5),
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        )

        used_datapipes["pv"] = pv_datapipe

    if use_topo:
        logger.debug("Opening Topographic Data")
        topo_datapipe = OpenTopography(configuration.input_data.topographic.topographic_filename)

        used_datapipes["topo"] = topo_datapipe

    return used_datapipes


def get_and_return_overlapping_time_periods_and_t0(used_datapipes: dict, key_for_t0: str = "gsp"):
    """
    Takes datapipes and obtains the overlapping time periods + t0 time datapipes

    Args:
        used_datapipes: Dictionary of datapipes to compute the time intersection of
        key_for_t0: Key to use for the t0 datapipe

    Returns:
        Dictionary of datapipes with the proper time slices selected
    """
    datapipes_for_time_periods = []  # Using later to compute intersections
    datapipes_to_return = {}  # Returned along with original ones
    t0_datapipe = None
    configuration = used_datapipes.pop("config")
    for key, datapipe in used_datapipes.items():
        if "topo" in key:
            continue
        if key_for_t0 in key:
            forked_datapipes = datapipe.fork(3, buffer_size=100)
            t0_datapipe = forked_datapipes[2]
        else:
            forked_datapipes = datapipe.fork(2, buffer_size=100)
        datapipes_to_return[key] = forked_datapipes[0]
        if "nwp" == key:
            time_periods_datapipe = forked_datapipes[1].get_contiguous_time_periods(
                sample_period_duration=timedelta(hours=3),  # Init times are 3 hours apart
                history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
                forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
                time_dim="init_time_utc",
            )
            datapipes_for_time_periods.append(time_periods_datapipe)

        if "sat" == key:
            time_periods_datapipe = forked_datapipes[1].get_contiguous_time_periods(
                sample_period_duration=timedelta(minutes=5),
                history_duration=timedelta(
                    minutes=configuration.input_data.satellite.history_minutes
                ),
                forecast_duration=timedelta(minutes=0),
            )
            datapipes_for_time_periods.append(time_periods_datapipe)

        if "hrv" == key:
            time_periods_datapipe = forked_datapipes[1].get_contiguous_time_periods(
                sample_period_duration=timedelta(minutes=5),
                history_duration=timedelta(
                    minutes=configuration.input_data.hrvsatellite.history_minutes
                ),
                forecast_duration=timedelta(minutes=0),
            )
            datapipes_for_time_periods.append(time_periods_datapipe)

        if "pv" == key:
            time_periods_datapipe = forked_datapipes[1].get_contiguous_time_periods(
                sample_period_duration=timedelta(minutes=5),
                history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
                forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
            )
            datapipes_for_time_periods.append(time_periods_datapipe)
        if "gsp" == key:
            time_periods_datapipe = forked_datapipes[1].get_contiguous_time_periods(
                sample_period_duration=timedelta(minutes=30),
                history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
                forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
            )
            datapipes_for_time_periods.append(time_periods_datapipe)

    # Now have the forked ones
    # find joint overlapping timer periods
    logger.debug("Getting joint time periods")
    overlapping_datapipe = datapipes_for_time_periods[0].select_overlapping_time_slice(
        secondary_datapipes=datapipes_for_time_periods[1:],
    )

    # select time periods
    t0_datapipe = t0_datapipe.select_time_periods(time_periods=overlapping_datapipe)

    num_t0_datapipes = len(datapipes_to_return.keys())  # One for each input
    t0_datapipes = t0_datapipe.select_t0_time(return_all_times=False).fork(
        num_t0_datapipes, buffer_size=100
    )

    for i, key in enumerate(list(datapipes_to_return.keys())):
        datapipes_to_return[key + "_t0"] = t0_datapipes[i]

    # Readd config for later
    datapipes_to_return["config"] = configuration
    if "topo" in used_datapipes.keys():
        datapipes_to_return["topo"] = used_datapipes["topo"]
    return datapipes_to_return


def add_selected_time_slices_from_datapipes(used_datapipes: dict):
    """
    Takes datapipes and t0 datapipes and returns the sliced datapipes

    Args:
        used_datapipes: Dictionary of used datapipes and t0 ones

    Returns:
        Dictionary of datapipes after the time slices are selected
    """
    datapipes_to_return = {}  # Returned along with original ones
    configuration = used_datapipes.pop("config")
    for key, datapipe in used_datapipes.items():
        if "topo" in key:
            continue
        if "_t0" in key:
            continue
        if "nwp" == key:
            datapipes_to_return[key] = datapipe.convert_to_nwp_target_time(
                t0_datapipe=used_datapipes[key + "_t0"],
                sample_period_duration=timedelta(hours=1),
                history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
                forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
            )

        if "sat" == key:
            datapipes_to_return[key] = datapipe.select_time_slice(
                t0_datapipe=used_datapipes[key + "_t0"],
                history_duration=timedelta(
                    minutes=configuration.input_data.satellite.history_minutes
                ),
                forecast_duration=timedelta(minutes=0),
                sample_period_duration=timedelta(minutes=5),
            )

        if "hrv" == key:
            datapipes_to_return[key] = datapipe.select_time_slice(
                t0_datapipe=used_datapipes[key + "_t0"],
                history_duration=timedelta(
                    minutes=configuration.input_data.hrvsatellite.history_minutes
                ),
                forecast_duration=timedelta(minutes=0),
                sample_period_duration=timedelta(minutes=5),
            )

        if "pv" == key:
            pv_1, pv_2 = used_datapipes[key + "_t0"].fork(2)
            pv_dp1, pv_dp2 = datapipe.fork(2)
            datapipes_to_return[key] = pv_dp1.select_time_slice(
                t0_datapipe=pv_1,
                history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
                forecast_duration=timedelta(minutes=0),
                sample_period_duration=timedelta(minutes=5),
            )
            datapipes_to_return[key + "_future"] = pv_dp2.select_time_slice(
                t0_datapipe=pv_2,
                history_duration=timedelta(minutes=0),
                forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
                sample_period_duration=timedelta(minutes=5),
            )

        if "gsp" == key:
            gsp_1, gsp_2 = used_datapipes[key + "_t0"].fork(2)
            gsp_dp1, gsp_dp2 = datapipe.fork(2)
            datapipes_to_return[key] = gsp_dp1.select_time_slice(
                t0_datapipe=gsp_1,
                history_duration=timedelta(minutes=configuration.input_data.gsp.history_minutes),
                forecast_duration=timedelta(minutes=0),
                sample_period_duration=timedelta(minutes=30),
            )
            datapipes_to_return[key + "_future"] = gsp_dp2.select_time_slice(
                t0_datapipe=gsp_2,
                history_duration=timedelta(minutes=0),
                forecast_duration=timedelta(minutes=configuration.input_data.gsp.forecast_minutes),
                sample_period_duration=timedelta(minutes=30),
            )
    if "topo" in used_datapipes.keys():
        datapipes_to_return["topo"] = used_datapipes["topo"]
    datapipes_to_return["config"] = configuration
    return datapipes_to_return


def create_t0_and_loc_datapipes(
    datapipes_dict: dict,
    configuration: Configuration,
    key_for_t0: str = "gsp",
    shuffle: bool = True,
    nwp_max_t0_offset: timedelta = timedelta(minutes=0),
):
    """
    Takes source datapipes and returns datapipes of appropriate sample pairs of locations and times.

    The (location, t0) pairs are sampled without replacement.

    Args:
        datapipes_dict: Dictionary of datapipes of input sources for which we want to select
            appropriate location and times.
        configuration: Configuration object for inputs.
        key_for_t0: Key to use for the t0 datapipe. Must be "gsp" or "pv".
        shuffle: Whether to use the internal shuffle function when yielding location times. Else
            location times will be heavily ordered.
        nwp_max_t0_offset: If using dropout on NWP, sometimes we have to go back to previous NWP
            init time. In order to accomodat for this possibility in selecting times, set
            `nwp_max_t0_offset` as the max NWP dropout delay you plan to use.

    Returns:
        location datapipe, t0 datapipe

    """
    assert key_for_t0 in datapipes_dict
    assert key_for_t0 in ["gsp", "pv"]

    contiguous_time_datapipes = []  # Used to store contiguous time periods from each data source

    datapipes_dict[key_for_t0], key_datapipe = datapipes_dict[key_for_t0].fork(2, buffer_size=5)

    for key in datapipes_dict.keys():
        if key in ["topo"]:
            continue

        elif key == "nwp":
            sample_frequency = 180  # Init times are 3 hours apart
            history_duration = configuration.input_data.nwp.history_minutes
            forecast_duration = configuration.input_data.nwp.forecast_minutes
            time_dim = "init_time_utc"
            max_t0_offset = nwp_max_t0_offset

        elif key == "sat":
            sample_frequency = 5
            history_duration = configuration.input_data.satellite.history_minutes
            forecast_duration = 0
            time_dim = "time_utc"
            max_t0_offset = timedelta(minutes=0)

        elif key == "hrv":
            sample_frequency = 5
            history_duration = configuration.input_data.hrvsatellite.history_minutes
            forecast_duration = 0
            time_dim = "time_utc"
            max_t0_offset = timedelta(minutes=0)

        elif key == "pv":
            sample_frequency = 5
            history_duration = configuration.input_data.pv.history_minutes
            forecast_duration = configuration.input_data.pv.forecast_minutes
            time_dim = "time_utc"
            max_t0_offset = timedelta(minutes=0)

        elif key == "gsp":
            sample_frequency = 30
            history_duration = configuration.input_data.gsp.history_minutes
            forecast_duration = configuration.input_data.gsp.forecast_minutes
            time_dim = "time_utc"
            max_t0_offset = timedelta(minutes=0)

        else:
            raise ValueError(f"Unexpected key: {key}")

        datapipes_dict[key], datapipe_copy = datapipes_dict[key].fork(2, buffer_size=5)

        time_periods = datapipe_copy.get_contiguous_time_periods(
            sample_period_duration=timedelta(minutes=sample_frequency),
            history_duration=timedelta(minutes=history_duration),
            forecast_duration=timedelta(minutes=forecast_duration),
            time_dim=time_dim,
            max_t0_offset=max_t0_offset,
        )

        contiguous_time_datapipes.append(time_periods)

    # Find joint overlapping contiguous time periods
    if len(contiguous_time_datapipes) > 1:
        logger.debug("Getting joint time periods")
        overlapping_datapipe = contiguous_time_datapipes[0].select_overlapping_time_slice(
            secondary_datapipes=contiguous_time_datapipes[1:],
        )
    else:
        logger.debug("Skipping getting joint time periods")
        overlapping_datapipe = contiguous_time_datapipes[0]

    # Select time periods and set length
    key_datapipe = key_datapipe.select_time_periods(time_periods=overlapping_datapipe)

    t0_loc_datapipe = key_datapipe.select_loc_and_t0(return_all=True, shuffle=shuffle)

    location_pipe, t0_datapipe = t0_loc_datapipe.unzip(sequence_length=2)

    return location_pipe, t0_datapipe



"""Fill nighttime PV with NaNs"""

import logging
from typing import Union

import numpy as np
import pvlib
import xarray as xr
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

from ocf_datapipes.utils.geospatial import osgb_to_lat_lon

logger = logging.getLogger(__name__)


@functional_datapipe("pv_fill_night_nans")
class PVFillNightNansIterDataPipe(IterDataPipe):
    """Fill nighttime nans with zeros"""

    def __init__(self, source_datapipe: IterDataPipe, elevation_limit: int = 5):
        """
        Fill nighttime nans with zeros

        Args:
            source_datapipe: the main data pipe
            elevation_limit: below this limit, all nans will be filled wiht zero.
                This is defaulted to to 5.
        """
        self.source_datapipe = source_datapipe
        self.elevation_limit = elevation_limit

    def __iter__(self) -> Union[xr.DataArray, xr.Dataset]:
        """Run iter"""

        for xr_data in self.source_datapipe:
            logger.info("Going to fill night time nans")

            lats, lons = osgb_to_lat_lon(x=xr_data.x_osgb, y=xr_data.y_osgb)

            elevation = np.full_like(xr_data.data, fill_value=np.NaN).astype(np.float32)
            for example_idx, (lat, lon) in enumerate(zip(lats, lons)):
                logger.debug(
                    f"Getting solar elevation for {lat} {lon} " f"{example_idx} out of {len(lats)}"
                )

                # get mask data for nans
                nan_mask = np.isnan(xr_data.data[:, example_idx])
                time_utc_nans = xr_data.time_utc[nan_mask]

                solpos = pvlib.solarposition.get_solarposition(
                    time=time_utc_nans,
                    latitude=lat,
                    longitude=lon,
                    # Which `method` to use?
                    # pyephem seemed to be a good mix between speed and ease but causes
                    # segfaults!
                    # nrel_numba doesn't work when using multiple worker processes.
                    # nrel_c is probably fastest but requires C code to be manually compiled:
                    # https://midcdmz.nrel.gov/spa/
                )
                elevation[nan_mask, example_idx] = solpos["elevation"]

            # get maks data for nighttime and nans
            night_time_mask = elevation < self.elevation_limit
            nan_mask = np.isnan(xr_data.data)
            total_mask = night_time_mask & nan_mask

            # set value
            logger.debug("Setting night nans to 0")
            xr_data.data[total_mask] = 0.0

            while True:
                yield xr_data



"""Wrapper for Power Perceiver Production Data Pipeline"""
import logging
from datetime import timedelta
from pathlib import Path
from typing import Optional, Union

import xarray
from torchdata.datapipes.iter import IterDataPipe

import ocf_datapipes  # noqa
from ocf_datapipes.batch import MergeNumpyModalities
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.load import OpenConfiguration, OpenGFSForecast, OpenNWPID, OpenPVFromNetCDF
from ocf_datapipes.utils.consts import NWP_GFS_MEAN, NWP_GFS_STD, NWP_MEAN, NWP_STD

logger = logging.getLogger(__name__)
xarray.set_options(keep_attrs=True)

# should scale with batch_size #TODO
BUFFER_SIZE = -1


def nwp_pv_datapipe(
    configuration_filename: Union[Path, str], tag: Optional[str] = "train"
) -> IterDataPipe:
    """
    Create the Power Perceiver production pipeline using a configuration

    Args:
        configuration_filename: Name of the configuration
        tag: Tag for train or test

    Returns:
        DataPipe ready to be put in a Dataloader for production
    """
    ####################################
    #
    # main data pipe for loading a simple site level forecast
    #
    #####################################
    # load configuration
    config_datapipe = OpenConfiguration(configuration_filename)
    configuration: Configuration = next(iter(config_datapipe))

    logger.debug("Opening Datasets")
    pv_datapipe, pv_location_datapipe = (
        OpenPVFromNetCDF(pv=configuration.input_data.pv)
        .pv_fill_night_nans()
        .fork(2, buffer_size=BUFFER_SIZE)
    )

    if configuration.input_data.nwp.nwp_provider == "UKMetOffice":
        nwp_datapipe = OpenNWPID(configuration.input_data.nwp.nwp_zarr_path)
    elif configuration.input_data.nwp.nwp_provider == "GFS":
        nwp_datapipe = OpenGFSForecast(configuration.input_data.nwp.nwp_zarr_path)
    else:
        raise Exception(
            f"NWP provider {configuration.input_data.nwp.nwp_provider} "
            f'not in "UKMetOffice" or "GFS"'
        )

    logger.debug("Add t0 idx and normalize")
    pv_datapipe = pv_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
    ).normalize(normalize_fn=lambda x: x / x.capacity_watt_power)
    nwp_datapipe = nwp_datapipe.add_t0_idx_and_sample_period_duration(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.nwp.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
    )

    if tag == "test":
        return_all = True
    else:
        return_all = False
    logger.debug("Getting locations")
    (
        location_datapipe1,
        location_datapipe2,
        location_datapipe3,
        location_datapipe4,
    ) = pv_location_datapipe.location_picker(
        return_all_locations=return_all,
        x_dim_name=configuration.input_data.pv.x_dim_name,
        y_dim_name=configuration.input_data.pv.y_dim_name,
    ).fork(
        4, buffer_size=BUFFER_SIZE
    )

    logger.debug("Making PV space slice")
    pv_datapipe = pv_datapipe.select_id(location_datapipe=location_datapipe1, data_source_name="pv")

    # filter out nans and remove zero data, do no this for testing
    if tag != "test":
        pv_datapipe = pv_datapipe.pv_remove_zero_data(
            window=timedelta(
                minutes=configuration.input_data.pv.history_minutes
                + configuration.input_data.pv.forecast_minutes
            )
        ).remove_nans()

    # split into 3 forks
    pv_datapipe, pv_time_periods_datapipe, pv_t0_datapipe = pv_datapipe.fork(
        3, buffer_size=BUFFER_SIZE
    )

    # select square from nwp data
    if configuration.input_data.nwp.index_by_id:
        nwp_datapipe = nwp_datapipe.select_id(
            location_datapipe=location_datapipe2,
        )
    else:
        nwp_datapipe = nwp_datapipe.select_spatial_slice_pixels(
            location_datapipe=location_datapipe2,
            roi_height_pixels=configuration.input_data.nwp.nwp_image_size_pixels_height,
            roi_width_pixels=configuration.input_data.nwp.nwp_image_size_pixels_width,
            y_dim_name=configuration.input_data.nwp.y_dim_name,
            x_dim_name=configuration.input_data.nwp.x_dim_name,
        )

    nwp_datapipe, nwp_time_periods_datapipe = nwp_datapipe.fork(2, buffer_size=BUFFER_SIZE)

    #
    # get contiguous time periods
    pv_time_periods_datapipe = pv_time_periods_datapipe.get_contiguous_time_periods(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.pv.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
    )

    nwp_time_periods_datapipe = nwp_time_periods_datapipe.get_contiguous_time_periods(
        sample_period_duration=timedelta(
            minutes=configuration.input_data.nwp.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
        time_dim="init_time_utc",
    )

    # find joint overlapping timer periods
    overlapping_datapipe = pv_time_periods_datapipe.select_overlapping_time_slice(
        secondary_datapipes=[nwp_time_periods_datapipe],
        location_datapipe=location_datapipe3,
    )

    # select time periods
    pv_t0_datapipe = pv_t0_datapipe.select_time_periods(time_periods=overlapping_datapipe)

    # select t0 periods
    location_datapipe4 = location_datapipe3.number_of_locations()
    pv_t0_datapipe, nwp_t0_datapipe = pv_t0_datapipe.select_t0_time(
        return_all_times=return_all, number_locations_datapipe=location_datapipe4
    ).fork(2)

    # take pv time slices
    pv_datapipe = (
        pv_datapipe.select_time_slice(
            t0_datapipe=pv_t0_datapipe,
            history_duration=timedelta(minutes=configuration.input_data.pv.history_minutes),
            forecast_duration=timedelta(minutes=configuration.input_data.pv.forecast_minutes),
            sample_period_duration=timedelta(
                minutes=configuration.input_data.pv.time_resolution_minutes
            ),
        )
        .convert_pv_to_numpy_batch()
        .merge_numpy_examples_to_batch(n_examples_per_batch=configuration.process.batch_size)
    )

    # take nwp time slices
    nwp_datapipe = nwp_datapipe.convert_to_nwp_target_time(
        t0_datapipe=nwp_t0_datapipe,
        sample_period_duration=timedelta(
            minutes=configuration.input_data.nwp.time_resolution_minutes
        ),
        history_duration=timedelta(minutes=configuration.input_data.nwp.history_minutes),
        forecast_duration=timedelta(minutes=configuration.input_data.nwp.forecast_minutes),
    )

    if configuration.input_data.nwp.nwp_provider == "UKMetOffice":
        nwp_datapipe = nwp_datapipe.normalize(mean=NWP_MEAN, std=NWP_STD)
    else:
        nwp_datapipe = nwp_datapipe.normalize(mean=NWP_GFS_MEAN, std=NWP_GFS_STD)

    nwp_datapipe = nwp_datapipe.convert_nwp_to_numpy_batch().merge_numpy_examples_to_batch(
        n_examples_per_batch=configuration.process.batch_size
    )

    ####################################
    #
    # Join data pipes together, and get extra details
    #
    #####################################
    logger.debug("Combine all the data sources")
    combined_datapipe = (
        MergeNumpyModalities([pv_datapipe, nwp_datapipe])
        .encode_space_time()
        .add_sun_position(modality_name="pv")
    )

    # combined_datapipe = combined_datapipe.add_length(
    #     configuration=configuration, train_validation_test=tag
    # )

    return combined_datapipe



# Copyright (c) Meta Platforms, Inc. and affiliates.
# All rights reserved.
#
# This source code is licensed under the BSD-style license found in the
# LICENSE file in the root directory of this source tree.

from collections import deque
from concurrent import futures
from typing import Callable, Iterator, Optional, Sized, TypeVar

from torch.utils.data.datapipes.utils.common import _check_unpickable_fn, validate_input_col
from torchdata.datapipes import functional_datapipe
from torchdata.datapipes.iter import IterDataPipe

T_co = TypeVar("T_co", covariant=True)


def _no_op_fn(*args):
    """
    No-operation function, returns passed arguments.
    """
    if len(args) == 1:
        return args[0]
    return args


@functional_datapipe("ocf_threadpool_map")
class ThreadPoolMapperIterDataPipe(IterDataPipe[T_co]):
    r"""
    Applies a function over each item from the source DataPipe concurrently
    using ``ThreadPoolExecutor`` (functional name: ``threadpool_map``).
    The function can be any regular Python function or partial object. Lambda
    function is not recommended as it is not supported by pickle.

    Args:
        source_datapipe: Source IterDataPipe
        fn: Function being applied over each item
        input_col: Index or indices of data which ``fn`` is applied, such as:
            - ``None`` as default to apply ``fn`` to the data directly.
            - Integer(s) is used for list/tuple.
            - Key(s) is used for dict.
        output_col: Index of data where result of ``fn`` is placed. ``output_col`` can be specified
            only when ``input_col`` is not ``None``
            - ``None`` as default to replace the index that ``input_col`` specified; For ``input_col`` with
              multiple indices, the left-most one is used, and other indices will be removed.
            - Integer is used for list/tuple. ``-1`` represents to append result at the end.
            - Key is used for dict. New key is acceptable.
        scheduled_tasks: How many tasks will be scheduled at any given time (Default value: 128)
        max_workers: Maximum number of threads to execute function calls
        **threadpool_kwargs: additional arguments to be given to the ``ThreadPoolExecutor``
    Note:
         For more information about ``max_workers`` and additional arguments for the ``ThreadPoolExecutor``
         please refer to: https://docs.python.org/3/library/concurrent.futures.html#concurrent.futures.ThreadPoolExecutor
    Note:
        For optimal use of all threads, ``scheduled_tasks`` > ``max_workers`` is strongly recommended. The higher the
        variance of the time needed to finish execution of the given ``fn`` is, the higher the value
        of ``scheduled_tasks`` needs to be to avoid threads sitting idle while waiting
        for the next result (as results are returned in correct order).
        However, too high value of ``scheduled_tasks`` might lead to long waiting period until the first element is yielded
        as ``next`` is called ``scheduled_tasks`` many times on ``source_datapipe`` before yielding.
        We encourage you to try out different values of ``max_workers`` and ``scheduled_tasks``
        in search for optimal values for your use-case.


    Example:
    .. testsetup::
        from torchdata.datapipes.iter import IterableWrapper
        import requests
        import time
        from unittest.mock import MagicMock
        requests.get = MagicMock()
        urls = []
    .. testcode::
        # fetching html from remote
        def fetch_html(url: str, **kwargs):
            r = requests.get(url, **kwargs)
            r.raise_for_status()
            return r.content
        dp = IterableWrapper(urls)
        dp = dp.threadpool_map(fetch_html,max_workers=16)
    .. testcode::
        def mul_ten(x):
            time.sleep(0.1)
            return x * 10
        dp = IterableWrapper([(i, i) for i in range(50)])
        dp = dp.threadpool_map(mul_ten, input_col=1)
        print(list(dp))
    .. testoutput::
        [(0, 0), (1, 10), (2, 20), (3, 30), ...]
    .. testcode::
        dp = IterableWrapper([(i, i) for i in range(50)])
        dp = dp.threadpool_map(mul_ten, input_col=1, output_col=-1)
        print(list(dp))
    .. testoutput::
        [(0, 0, 0), (1, 1, 10), (2, 2, 20), (3, 3, 30), ...]
    """

    datapipe: IterDataPipe
    fn: Callable

    def __init__(
        self,
        source_datapipe: IterDataPipe,
        fn: Callable,
        input_col=None,
        output_col=None,
        scheduled_tasks: int = 128,
        max_workers: Optional[int] = None,
        **threadpool_kwargs,
    ) -> None:
        super().__init__()
        self.datapipe = source_datapipe

        _check_unpickable_fn(fn)
        self.fn = fn  # type: ignore[assignment]

        if scheduled_tasks <= 0:
            raise ValueError("'scheduled_tasks' is required to be a positive integer.")
        self.scheduled_tasks = scheduled_tasks
        if max_workers is not None and max_workers <= 0:
            raise ValueError("'max_workers' is required to be a positive integer.")
        self.max_workers = max_workers
        self.threadpool_kwargs = threadpool_kwargs

        self.input_col = input_col
        if input_col is None and output_col is not None:
            raise ValueError("`output_col` must be None when `input_col` is None.")
        if isinstance(output_col, (list, tuple)):
            if len(output_col) > 1:
                raise ValueError("`output_col` must be a single-element list or tuple")
            output_col = output_col[0]
        self.output_col = output_col
        validate_input_col(fn, input_col)

    def _apply_fn(self, data):
        if self.input_col is None and self.output_col is None:
            return self.fn(data)

        if self.input_col is None:
            res = self.fn(data)
        elif isinstance(self.input_col, (list, tuple)):
            args = tuple(data[col] for col in self.input_col)
            res = self.fn(*args)
        else:
            res = self.fn(data[self.input_col])

        # Copy tuple to list and run in-place modification because tuple is immutable.
        if isinstance(data, tuple):
            t_flag = True
            data = list(data)
        else:
            t_flag = False

        if self.output_col is None:
            if isinstance(self.input_col, (list, tuple)):
                data[self.input_col[0]] = res
                for idx in sorted(self.input_col[1:], reverse=True):
                    del data[idx]
            else:
                data[self.input_col] = res
        else:
            if self.output_col == -1:
                data.append(res)
            else:
                data[self.output_col] = res

        # Convert list back to tuple
        return tuple(data) if t_flag else data

    def __iter__(self) -> Iterator[T_co]:
        with futures.ThreadPoolExecutor(
            max_workers=self.max_workers, **self.threadpool_kwargs
        ) as executor:
            futures_deque: deque = deque()
            has_next = True
            itr = iter(self.datapipe)
            for _ in range(self.scheduled_tasks):
                try:
                    futures_deque.append(executor.submit(self._apply_fn, next(itr)))
                except StopIteration:
                    has_next = False
                    break

            while len(futures_deque) > 0:
                if has_next:
                    try:
                        futures_deque.append(executor.submit(self._apply_fn, next(itr)))
                    except StopIteration:
                        has_next = False
                yield futures_deque.popleft().result()

    def __len__(self) -> int:
        if isinstance(self.datapipe, Sized):
            return len(self.datapipe)
        raise TypeError(f"{type(self).__name__} instance doesn't have valid length")



""" Make fake batch """
from datetime import datetime, timezone
from typing import Optional, Union

import numpy as np
import torch
from torchdata.datapipes.iter import IterableWrapper

from ocf_datapipes.batch.fake.gsp import make_fake_gsp_data
from ocf_datapipes.batch.fake.nwp import make_fake_nwp_data
from ocf_datapipes.batch.fake.pv import make_fake_pv_data
from ocf_datapipes.batch.fake.satellite import make_fake_satellite_data
from ocf_datapipes.batch.fake.sun import make_fake_sun_data
from ocf_datapipes.config.load import load_yaml_configuration
from ocf_datapipes.config.model import Configuration
from ocf_datapipes.utils.utils import datetime64_to_float


def make_fake_batch(configuration: Configuration, to_torch: Optional[bool] = False) -> dict:
    """
    Make a random fake batch, this is useful for models that use this object

    Args:
        configuration: a configuration file
        to_torch: optional if we return the batch with torch.Tensor

    Returns: dictionary containing the batch

    """

    # time now rounded down to 5 minutes
    t0_datetime_utc = datetime.now(tz=timezone.utc)
    t0_datetime_utc = t0_datetime_utc.replace(minute=t0_datetime_utc.minute // 5 * 5)
    t0_datetime_utc = t0_datetime_utc.replace(second=0)
    t0_datetime_utc = t0_datetime_utc.replace(microsecond=0)

    # make fake PV data
    batch_pv = make_fake_pv_data(configuration=configuration, t0_datetime_utc=t0_datetime_utc)

    # make NWP data
    batch_nwp = make_fake_nwp_data(configuration=configuration, t0_datetime_utc=t0_datetime_utc)

    # make GSP data
    batch_gsp = make_fake_gsp_data(configuration=configuration, t0_datetime_utc=t0_datetime_utc)

    # make hrv and normal satellite data
    batch_satellite = make_fake_satellite_data(
        configuration=configuration, t0_datetime_utc=t0_datetime_utc, is_hrv=False
    )
    batch_hrv_satellite = make_fake_satellite_data(
        configuration=configuration, t0_datetime_utc=t0_datetime_utc, is_hrv=True
    )

    # make sun features
    batch_sun = make_fake_sun_data(configuration=configuration)

    batch = {
        **batch_pv,
        **batch_nwp,
        **batch_gsp,
        **batch_satellite,
        **batch_hrv_satellite,
        **batch_sun,
    }

    if to_torch:
        for k, v in batch.items():
            if isinstance(v, int):
                batch[k] = torch.IntTensor(v)
            elif isinstance(v, np.ndarray):
                if v.dtype == "datetime64[s]":
                    batch[k] = torch.from_numpy(datetime64_to_float(v))
                else:
                    batch[k] = torch.from_numpy(v)

    return batch


def fake_data_pipeline(configuration: Union[str, Configuration]):
    """
    Make a fake data pipeline

    Args:
        configuration: a configuration file

    """

    if isinstance(configuration, str):
        configuration = load_yaml_configuration(configuration)

    batch = make_fake_batch(configuration=configuration, to_torch=True)

    def fake_iter():
        while True:
            yield batch

    return IterableWrapper(fake_iter())
